{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjArray.Pas
|
| Description: Array container.
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| History: Version 1.0  March 1996. 
|          May 14, 1997. MB add CompareFunc property
|
|---------------------------------------------------------------------------}
unit ObjArray;

{$I Contain.Inc}

interface

uses
    ObjPerst, { TObjectReader, TObjectWriter }
    ObjCmp,   { TCompareFunc }
    ObjType,  { TArrayIndex }
    ObjBase;  { TContainer }


type

TpObjArray = ^TObjArray;
TObjArray = array[1..CMaxArrayIndex] of TObject;

TArray = class(TContainer)
protected
    { number of objects the array can hold. The base class FSize data
      member holds the current number of objects. }
    FCapacity : TArrayIndex;

    { the number of FCapacity elements that will be increased when Insert is
      invoked and FSize has reached FCapacity. If FDelta is 0, then container
      will not increase its Capacity, and raise EContainerIsFull on Insert }
    FDelta    : TArrayIndex;

    { True if array is currently sorted }
    FSorted   : Boolean;

    { returns an object at a specific index. May raise EInvalidIndex.
      It is the programmer's responsibility not to change the sort order,
      when a call to BinarySearch is to be made. Here is the potential
      problem:
      1. Invoke XXXSort  2. Read an object from the ObjArray property.
      3. Modify the object read in a manner that changes the sort order
      4. Invoke BinarySearch when the array is not sorted any more }
    function  GetObj(const Index : TArrayIndex) : TObject;

    { Sets the object at a specfic index. The current object is Freed. The
      Sorted flag may be changed. May raise EInvalidIndex. }
    procedure SetObj(const Index : TArrayIndex; const Obj : TObject); virtual;

    { a new array is allocated, and all the objects are copied to the new
      array. May raise ENewArraySizeTooSmall if InCapacity is too small for
      the current contents. }
    procedure SetCapacity(const InCapacity : TArrayIndex);

    { sets a new value for FDelta. }
    procedure SetDelta(const InDelta : TArrayIndex);

    { returns the number of free cells in the array }
    function GetFreeSpace : TArrayIndex;

    procedure SetCompareFunc( const InCompareFunc : TCompareFunc );

public
    { Capacity, is the initial capacity of the array. After the number of
      objects inserted reached Capacity, the array capacity is increased
      by Delta }
    constructor Create( const InObjectType  : TClass;
                        const InCompareFunc  : TCompareFunc;
                        const InCapacity     : TArrayIndex;
                        const InDelta        : TArrayIndex );

    { destroy invokes Clear to remove and free every element in the container }
    destructor Destroy; override;

    { the Default array property may be used to examine an element or change
      an element. To add an object, always use the Insert method.
      May raise EInvalidIndex if Index < 1 or Index > FSize }
    property ObjArray[const Index : TArrayIndex] : TObject read  GetObj
                                                           write SetObj; default;

    { read only property which determines if array is currently sorted. To sort
      the array, use QuickSort or ShellSort. }
    property Sorted : Boolean read FSorted;

    { inserts an object in the next available index. May raise
      EContainerFull if FDelta = 0. The Sorted state may be affected }
    procedure Insert(const Obj : TObject);

    { swaps two objects in the array. May raise EInvalidIndex. May change
      the Sorted status of the array if the array was previously sorted }
    procedure Swap(const Index1,Index2 : TArrayIndex);

    { Deletes (removes and Frees) the object at Index. The last element is
      moved at Index. The Sorted flag may be affected. May raise
      EInvalidIndex.  }
    procedure Delete( const Index : TArrayIndex );

    { changing the comparison function changed sorted status to False,
      assuming that if the compare function is changed, the ordering is
      changed. }
    property CompareFunc : TCompareFunc read FCompareFunc
                                        write SetCompareFunc;

    { for each item in the container, call ProcIterator. ProcIterator is
      a file level procedure. }
    procedure ForEachCallProc( const ProcIterator : TProcIterator;
                               const CallBackData : LongInt); override;

    { for each item in the container, call MethodIterator. MethodIterator is
      a method of a class. }
    procedure ForEachCallMethod( const MethodIterator : TMethodIterator;
                                 const CallBackData   : LongInt); override;

    { sorts the array using quick sort. Very good performance when the
      array is quite random }
    procedure QuickSort;

    { sorts the array using the Shell-Metzner sort. Very good when the array
      is nearly sorted. For example, after a call to Delete. }
    procedure ShellSort;

    { comparative bench marks for QuickSort and ShellSort.
      On a random array:        QuickSort     ShellSort
                   1000 items   < 1 second   < 1 second
                   16000 items    7 seconds  35 seconds
      On a nearly ordered
      array:       16000 items   25 seconds  7 seconds
    }

    { performs a binary search and returns the index of the sought element.
      It is the programmers responsibility to ensure the array is currently
      sorted, by a previous call to SortXXXX
      Returns CInvalidArrayIndex (0) if not found.
      May raise EArrayNotSorted if Sorted is currently False }
    function BinarySearch(const Obj : TObject) : TArrayIndex;

    { performs a linear search and returns the index of the sought element
      Returns CInvalidArrayIndex (0) if not found. }
    function LinearSearch(const Obj : TObject) : TArrayIndex;

    { clear all the items in the container. Free is called for each item }
    procedure Clear; override;

    { Writing to Capacity, will resize the array }
    property Capacity : TArrayIndex read FCapacity
                                    write SetCapacity;

    property Delta : TArrayIndex read FDelta
                                 write SetDelta;

    { returns the number of free cells in the array }
    property FreeSpace : TArrayIndex read GetFreeSpace;

protected
    { pointer to array }
    pObjArray : TpObjArray;

    { swaps two objects. No range checking is performed }
    procedure LocalSwap(const Index1,Index2 : TArrayIndex);

    { raises EInvalidIndex if Index is not valid (not within 1..FSize) }
    procedure CheckIndex(const Index : TArrayIndex);

    { verifies if the array is still sorted after a modification at
      Index }
    procedure CheckSortedState(const Index : TArrayIndex);

end;

{-------------------------------------------------------------------------}

{ Streaming routines }

procedure ReadArray( const Obj    : TObject;
                     const Reader : TObjectReader ); far;

procedure WriteArray( const Obj    : TObject;
                      const Writer : TObjectWriter ); far;

{-------------------------------------------------------------------------}

implementation

uses
       ObjBuckt,  { TBucket }
       ObjStack;  { TStack }

{-------------------------------------------------------------------------}

constructor TArray.Create( const InObjectType   : TClass;
                           const InCompareFunc  : TCompareFunc;
                           const InCapacity     : TArrayIndex;
                           const InDelta        : TArrayIndex );
begin
    if InCapacity < 1 then
        RaiseInvalidArrayCapacity;

    FCompareFunc       := InCompareFunc;
    FCapacity          := InCapacity;
    FDelta             := InDelta;
    FSorted            := True; { an empty array is sorted }

    { in Delphi 1.0, an EIntOverflow error was raised with FCapacity > 8500,
      so a typecast to LongInt prevents the overflow }
    GetMem(pObjArray, LongInt(FCapacity)*SizeOf(TObject));

    try
        inherited Create(InObjectType);
    except
        { in Delphi 1.0, an EIntOverflow error was raised with FCapacity > 8500,
          so a typecast to LongInt prevents the overflow }
        FreeMem(pObjArray, LongInt(FCapacity)*SizeOf(TObject));
        Raise;
    end;
end;

{-------------------------------------------------------------------------}

destructor TArray.Destroy;
begin
    { call inherited Destroy to clear the container first }
    inherited Destroy;

    { in Delphi 1.0, an EIntOverflow error was raised with FCapacity > 8500,
      so a typecast to LongInt prevents the overflow }
    FreeMem(pObjArray, LongInt(FCapacity)*SizeOf(TObject));
end;

{-------------------------------------------------------------------------}

function  TArray.GetObj(const Index : TArrayIndex) : TObject;
begin
    CheckIndex(Index);

    Result := pObjArray^[Index];
end;

{-------------------------------------------------------------------------}

procedure TArray.SetObj(const Index : TArrayIndex; const Obj : TObject);
begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    CheckIndex(Index);

    { release the current object }
    ReleaseObj(pObjArray^[Index]);

    pObjArray^[Index] := Obj;

    CheckSortedState(Index);
end;

{-------------------------------------------------------------------------}

procedure TArray.SetCapacity(const InCapacity : TArrayIndex);
var
    pNewArray : TpObjArray;

    I         : TArrayIndex;

begin
    if InCapacity < 1 then
        RaiseInvalidArrayCapacity;

    if InCapacity = FCapacity then
        Exit; { no need to resize }

    if InCapacity < FSize then
        RaiseNewArraySizeTooSmall;

    { in Delphi 1.0, an EIntOverflow error was raised with FCapacity > 8500,
      so a typecast to LongInt prevents the overflow }
    GetMem(pNewArray, LongInt(InCapacity)*SizeOf(TObject));

    for I := 1 to FSize do
        pNewArray^[I] := pObjArray^[I];

    { in Delphi 1.0, an EIntOverflow error was raised with FCapacity > 8500,
      so a typecast to LongInt prevents the overflow }
    FreeMem(pObjArray, LongInt(FCapacity)*SizeOf(TObject));

    pObjArray := pNewArray;
    FCapacity := InCapacity;
end;

{-------------------------------------------------------------------------}

procedure TArray.SetDelta(const InDelta : TArrayIndex);
begin
    FDelta := InDelta;
end;

{-------------------------------------------------------------------------}

function TArray.GetFreeSpace : TArrayIndex;
begin
    Result := FCapacity - FSize;
end;

{-------------------------------------------------------------------------}

procedure TArray.CheckIndex(const Index : TArrayIndex);
begin
    if (Index < 1) or (Index > FSize) then
        RaiseInvalidIndex;
end;

{-------------------------------------------------------------------------}

procedure TArray.CheckSortedState(const Index : TArrayIndex);
begin
    { must have at least 2 objects to check }
    if FSize < 2 then
        Exit;

    if FSorted then
    begin
        if (Index = FSize) then
        begin
            { Index is last element, check previous }
            FSorted := not (FCompareFunc(pObjArray^[Index-1], pObjArray^[Index]) = cmpGreater);
        end
        else if (Index = 1) then
        begin
            { Index is first element, check next }
            FSorted := not (FCompareFunc(pObjArray^[Index], pObjArray^[Index+1]) = cmpGreater);
        end
        else
        begin
            { must check previous and next indexes }
            FSorted := not (FCompareFunc(pObjArray^[Index-1], pObjArray^[Index]) = cmpGreater) and
                       not (FCompareFunc(pObjArray^[Index], pObjArray^[Index+1]) = cmpGreater);
        end;
    end;
end;

{-------------------------------------------------------------------------}

procedure TArray.Insert(const Obj : TObject);
var
    NewCapacity : LongInt;
begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    if FSize = CMaxArrayIndex then
        RaiseContainerIsFull;   { cannot increase capacity }

    if (FSize >= FCapacity) then
    begin
        if FDelta = 0 then
            RaiseContainerIsFull; { fixed size container }

        NewCapacity := FCapacity + FDelta;
        if NewCapacity > CMaxArrayIndex then
            NewCapacity := CMaxArrayIndex;

        Capacity := NewCapacity;
    end;

    Inc(FSize);

    pObjArray^[FSize] := Obj;

    CheckSortedState(FSize);
end;

{-------------------------------------------------------------------------}

procedure TArray.Delete( const Index : TArrayIndex );
begin
    CheckIndex(Index);

    { release the object to delete }
    ReleaseObj(pObjArray^[Index]);

    if Index < FSize then
    begin
        { replace the deleted element by the last element }
        pObjArray^[Index] := pObjArray^[FSize];

        Dec(FSize);

        CheckSortedState(Index);
    end
    else
        { last element deleted, just adjust FSize }
        Dec(FSize);

    if FSize < 2 then
        FSorted := True;

end;

{-----------------------------------------------------------------------}

procedure TArray.SetCompareFunc( const InCompareFunc : TCompareFunc );
begin
    FCompareFunc := InCompareFunc;

    FSorted      := False;
end;

{-----------------------------------------------------------------------}

procedure TArray.ForEachCallProc( const ProcIterator : TProcIterator;
                                  const CallBackData : LongInt);
var
    I : TArrayIndex;
begin
    for I := 1 to FSize do
    begin
        ProcIterator(pObjArray^[I], CallBackData);
    end;
end;

{-----------------------------------------------------------------------}

procedure TArray.ForEachCallMethod( const MethodIterator : TMethodIterator;
                                    const CallBackData   : LongInt);
var
    I : TArrayIndex;
begin
    for I := 1 to FSize do
    begin
        MethodIterator(pObjArray^[I], CallBackData);
    end;
end;

{-------------------------------------------------------------------------}

function TArray.BinarySearch(const Obj : TObject) : TArrayIndex;
var
    High, Low, Median : TArrayIndex;
    Comparison        : TComparison;
    Found             : Boolean;

begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    if not Sorted and (FSize > 0) then
        RaiseArrayNotSorted;

    Median   := 1;   { prevent compiler warning }

    Found    := False;
    Low      := 1;
    High     := FSize;

    while (Low <= High) and not Found do
    begin
        Median := (Low + High) div 2;

        Comparison := FCompareFunc(Obj, pObjArray^[Median]);

        { check for range error at upper bound }
        if Median = CMaxArrayIndex then
        begin
            Found := (Comparison = cmpEqual);
            break; { exit loop to prevent Low := Median + 1 }
        end;

        if Comparison = cmpLess then
            High := Median - 1
        else if Comparison = cmpGreater then
            Low := Median + 1
        else
            Found := True;
    end;

    if not Found then
        BinarySearch := CInvalidArrayIndex
    else
	BinarySearch := Median;
end;

{-------------------------------------------------------------------------}

function TArray.LinearSearch(const Obj : TObject) : TArrayIndex;
var
    I     : TArrayIndex;
    Found : Boolean;

begin
    {$ifndef __RemoveTypeSafety}
    CheckType(Obj);
    {$endif}

    Found := False;
    I     := 0;

    if FSize > 0 then
    repeat
        Inc(I);

        if FCompareFunc(pObjArray^[I], Obj) = cmpEqual then
            Found := True;

    until Found or (I = FSize);

    if not Found then
        LinearSearch := CInvalidArrayIndex
    else
	LinearSearch := I;

end;

{-------------------------------------------------------------------------}

procedure TArray.Clear;
var
    I   : TArrayIndex;
begin
    for I := 1 to FSize do
        ReleaseObj(pObjArray^[I]);

    FSorted := True;

    FSize   := 0;
end;

{-------------------------------------------------------------------------}

procedure TArray.Swap(const Index1,Index2 : TArrayIndex);
begin
    CheckIndex(Index1);
    CheckIndex(Index2);

    LocalSwap(Index1, Index2);

    if FSorted then
    begin
        { sorted is only maintained if the values exchanged were identical }
        if FCompareFunc(pObjArray^[Index1], pObjArray^[Index2]) <> cmpEqual then
            FSorted := False;
    end;
end;

{-------------------------------------------------------------------------}

procedure TArray.LocalSwap(const Index1,Index2 : TArrayIndex);
var
    TempObj : TObject;

begin
    TempObj := pObjArray^[Index1];
    pObjArray^[Index1] := pObjArray^[Index2];
    pObjArray^[Index2] := TempObj;
end;

{-------------------------------------------------------------------------}

procedure TArray.ShellSort;
var
    Finished : Boolean;
    Jump     : TArrayIndex;
    I,J      : TArrayIndex;
begin

    Jump := FSize;

    while Jump > 1 do
    begin
        Jump := Jump div 2;
        repeat
            Finished := True;
            for J := 1 to (FSize - Jump) do
            begin
                I := J + Jump;
                if FCompareFunc(pObjArray^[J], pObjArray^[I]) = cmpGreater then
                begin
                    LocalSwap(J, I);
                    Finished := False;
                end;
            end;
        until Finished;
    end;

    FSorted := True;

end;

{-------------------------------------------------------------------------}

procedure TArray.QuickSort;

    type
        TLimits = record
            I, J : TArrayIndex;
        end;

var
    LimitStack : TStack;

    Bucket     : TBucket;
    Limits     : TLimits;

    I, J, Mid  : TArrayIndex;

    Pivot      : TObject;

    procedure AdjustPivot;
    begin
        { if the array has more than 5 objects adjust the pivot }
        if (J - I) > 5 then
        begin
            if ( (FCompareFunc(pObjArray^[Mid], Pivot) = cmpLess) and
                 (FCompareFunc(pObjArray^[Mid], pObjArray^[I]) = cmpGreater)
                ) or
               ( (FCompareFunc(pObjArray^[Mid], Pivot) = cmpGreater) and
                 (FCompareFunc(pObjArray^[Mid], pObjArray^[I]) = cmpLess)
                ) then
                LocalSwap(Mid, J)
            else if ( (FCompareFunc(pObjArray^[I], pObjArray^[Mid]) = cmpLess) and
                      (FCompareFunc(pObjArray^[I], Pivot) = cmpGreater)
                     ) or
                    ( (FCompareFunc(pObjArray^[I], pObjArray^[Mid]) = cmpGreater) and
                     (FCompareFunc(pObjArray^[I], Pivot) = cmpLess)
                     ) then
                LocalSwap(I, J)
        end;
    end;

begin
    LimitStack := TStack.Create(TBucket);

    try
        Limits.I := 1; Limits.J := FSize;
        Bucket := TBucket.CreateSetData(SizeOf(Limits), Limits);
        LimitStack.Push(Bucket);

        while LimitStack.Size > 0 do
        begin
            if TLimits((LimitStack.PeekTop as TBucket).DataPtr^).I >=
               TLimits((LimitStack.PeekTop as TBucket).DataPtr^).J then
                LimitStack.Pop.Free
            else
            begin
                (LimitStack.PeekTop as TBucket).GetData(Limits);
                I := Limits.I;
                J := Limits.J;
                Pivot := pObjArray^[J];
                Mid   := (I + J) div 2;
                AdjustPivot;
                Pivot := pObjArray^[J]; { set the pivot again }

                while I < J do
                begin
                    while FCompareFunc(pObjArray^[I], Pivot) = cmpLess do
                        Inc(I);
                    Dec(J);
                    while (I < J) and
                          (FCompareFunc(Pivot, pObjArray^[J]) = cmpLess) do
                        Dec(J);

                    if I < J then
                        LocalSwap(I, J);
                end; { while I < J }

                J := Limits.J;
                LocalSwap(I,J);

                if (I-Limits.I) >= (Limits.J - I) then
                begin
                    Limits.J := I - 1;
                    TLimits((LimitStack.PeekTop as TBucket).DataPtr^).I := I + 1;
                end
                else
                begin
                    Limits.I := I + 1;
                    TLimits((LimitStack.PeekTop as TBucket).DataPtr^).J := I - 1;
                end;

                LimitStack.Push( TBucket.CreateSetData(SizeOf(Limits), Limits));
            end; { if }
        end; { while }

    finally
        LimitStack.Free;
    end;

    FSorted := True;
end;

{-------------------------------------------------------------------------}

{ Streaming routines }

procedure ReadArray( const Obj    : TObject;
                     const Reader : TObjectReader );
var
    ObjArray : TArray;
    Size     : LongInt;
    Index    : LongInt;
    Item     : TObject;
begin
    ObjArray := Obj as TArray;

    ReadContainer(Obj, Reader );

    Size := Reader.ReadInteger;

    ObjArray.FCapacity := Reader.ReadInteger;

    ObjArray.FDelta    := Reader.ReadInteger;

    ObjArray.FSorted   := Reader.ReadBoolean;

    GetMem(ObjArray.pObjArray, LongInt(ObjArray.FCapacity)*SizeOf(TObject));

    for Index := 1 to Size do
    begin
        Item := Reader.ReadChildObject;

        ObjArray.Insert(Item);
    end;
end;

{-------------------------------------------------------------------------}

procedure WriteArray( const Obj    : TObject;
                      const Writer : TObjectWriter );
var
    Index    : LongInt;

    ObjArray : TArray;
begin
    ObjArray := Obj as TArray;

    WriteContainer(ObjArray, Writer);

    Writer.WriteInteger( ObjArray.Size );

    Writer.WriteInteger( ObjArray.Capacity );

    Writer.WriteInteger( ObjArray.Delta );

    Writer.WriteBoolean( ObjArray.Sorted );

    for Index := 1 to ObjArray.Size do
        Writer.WriteChildObject( ObjArray[Index] );
end;


{-------------------------------------------------------------------------}

initialization

    {$ifdef Win32}
    RegisterStreamable( TArray,
                        [TObjectReadProc(@ReadArray)],
                        [TObjectWriteProc(@WriteArray)] );
    {$else}
    RegisterStreamable( TArray,
                        [ReadArray],
                        [WriteArray] );
    {$endif}

end.
