{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjBuckt.Pas
|
| Description: Bucket objects to store non-object data in TContainer
|              derived classes
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| CAUTION: When streaming TIntegerBucket's and TShortIntBucket's be careful
|          with interoperability between 16 and 32 bit versions of Delphi,
|          as the data types are not the same size.
|
| History: Version 1.0  March 1996.
|                       May 07, 1997. Add stream procedures.
|          Sep 06, 1997. MB add CompareStringComboNoCase
|          May 23, 1998. MB add OwnerShip property to TStringCombo
|          Jul 06, 1998. MB use string pointers in TStringCombo to reduce
|                        memory load.
|          Jul 09, 1998. MB add hash functions for most buckets.
|          Apr 10, 1999. MB add TCurrencyBucket for 32 bit Delphi versions
|          Feb 05, 2001. MB make TStringCombo.Str writable
|
|---------------------------------------------------------------------------}
unit ObjBuckt;

{$I Contain.Inc}

interface

uses
    ObjType,  { TArrayIndex }
    ObjPerst, { TObjectReader, TObjectWriter }
    ObjCmp;   { TComparison }

type

{ class type which contains a string and a related object. Can be used
  in any TContainer. For an example, see the TStringTable container }
TStringCombo = class(TObject)
private
    FOwnership : Boolean;
    FStr       : String;
    FObj       : TObject;

    function GetObj    : TObject;

public
    constructor Create(const Str : String; const Obj : TObject);
    destructor Destroy; override;

    property Str : String read FStr write FStr;
    property Obj  : TObject read GetObj;
    property Ownership : Boolean read FOwnership write FOwnership;
end;

{ compare functions for TStringCombo objects }
function CompareStringCombo(const Lhs, Rhs : TObject) : TComparison; far;
function CompareStringComboNoCase(const Lhs, Rhs : TObject) : TComparison; far;

{ Hash function for TStringCombo class using StringHash }
function HashStringCombo( const Obj       : TObject;
                          const TableSize : TArrayIndex
                        ) : TArrayIndex; far;

{ Hash function for TStringCombo class using StringHash2 }
function HashStringCombo2( const Obj       : TObject;
                           const TableSize : TArrayIndex
                         ) : TArrayIndex; far;

{ streaming procedures }
procedure ReadStringCombo( const Obj     : TObject;
                           const Reader  : TObjectReader ); far;
procedure WriteStringCombo( const Obj     : TObject;
                            const Writer  : TObjectWriter ); far;

type

{ simple class to allow storing of any data into containers. For example,
  record instances may be stored inside a TBucket to be inserted into
  container classes. }
TBucket = class(TObject)
private
    FDataPtr    : Pointer;
    FDataSize   : Word;

public
    { InDataSize is in Bytes }
    constructor Create(const InDataSize : Word);
    constructor CreateSetData(const InDataSize : Word; const Data);
    destructor Destroy; override;

    { sets the data using an untyped reference }
    procedure SetData(const Data);

    { returns the data in an untyped reference }
    procedure GetData(var Data);

    { access the data as a generic pointer }
    property DataPtr : Pointer read FDataPtr;

end;

{ streaming procedures }
procedure ReadBucket( const Obj     : TObject;
                      const Reader  : TObjectReader ); far;
procedure WriteBucket( const Obj     : TObject;
                       const Writer  : TObjectWriter ); far;

{--------------------------------------------------------------------------}

type

TShortIntBucket = class(TBucket)
private
    function GetShortInt : ShortInt;
public
    constructor Create(const Data : ShortInt);
    property Value : ShortInt read GetShortInt;
end;

function CompareShortIntBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashShortIntBucket( const Obj       : TObject;
                             const TableSize : TArrayIndex
                            ) : TArrayIndex; far;



{--------------------------------------------------------------------------}

type

TIntegerBucket = class(TBucket)
private
    function GetInteger : Integer;
public
    constructor Create(const Data : Integer);
    property Value : Integer read GetInteger;
end;

function CompareIntegerBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashIntegerBucket( const Obj       : TObject;
                            const TableSize : TArrayIndex
                           ) : TArrayIndex; far;

{--------------------------------------------------------------------------}

type

TLongIntBucket = class(TBucket)
private
    function GetLongInt : LongInt;
public
    constructor Create(const Data : LongInt);
    property Value : LongInt read GetLongInt;
end;

function CompareLongIntBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashLongIntBucket( const Obj       : TObject;
                            const TableSize : TArrayIndex
                           ) : TArrayIndex; far;

{--------------------------------------------------------------------------}

type

TByteBucket = class(TBucket)
private
    function GetByte : Byte;
public
    constructor Create(const Data : Byte);
    property Value : Byte read GetByte;
end;

function CompareByteBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashByteBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex; far;

{--------------------------------------------------------------------------}

type

TWordBucket = class(TBucket)
private
    function GetWord : Word;
public
    constructor Create(const Data : Word);
    property Value : Word read GetWord;
end;

function CompareWordBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashWordBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex; far;

{--------------------------------------------------------------------------}

type

TCharBucket = class(TBucket)
private
    function GetChar : Char;
public
    constructor Create(const Data : Char);
    property Value : Char read GetChar;
end;

function CompareCharBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashCharBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex; far;

{--------------------------------------------------------------------------}

type

TBooleanBucket = class(TBucket)
private
    function GetBoolean : Boolean;
public
    constructor Create(const Data : Boolean);
    property Value : Boolean read GetBoolean;
end;

function CompareBooleanBucket(const Lhs, Rhs : TObject) : TComparison; far;

{--------------------------------------------------------------------------}

type

TWordBoolBucket = class(TBucket)
private
    function GetWordBool : WordBool;
public
    constructor Create(const Data : WordBool);
    property Value : WordBool read GetWordBool;
end;

function CompareWordBoolBucket(const Lhs, Rhs : TObject) : TComparison; far;

{--------------------------------------------------------------------------}

type

TLongBoolBucket = class(TBucket)
private
    function GetLongBool : LongBool;
public
    constructor Create(const Data : LongBool);
    property Value : LongBool read GetLongBool;
end;

function CompareLongBoolBucket(const Lhs, Rhs : TObject) : TComparison; far;

{--------------------------------------------------------------------------}

type

TRealBucket = class(TBucket)
private
    function GetReal : Real;
public
    constructor Create(const Data : Real);
    property Value : Real read GetReal;
end;

function CompareRealBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashRealBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex; far;


{--------------------------------------------------------------------------}

type

TSingleBucket = class(TBucket)
private
    function GetSingle : Single;
public
    constructor Create(const Data : Single);
    property Value : Single read GetSingle;
end;

function CompareSingleBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashSingleBucket( const Obj       : TObject;
                           const TableSize : TArrayIndex
                          ) : TArrayIndex; far;

{--------------------------------------------------------------------------}

type

TDoubleBucket = class(TBucket)
private
    function GetDouble : Double;
public
    constructor Create(const Data : Double);
    property Value : Double read GetDouble;
end;

function CompareDoubleBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashDoubleBucket( const Obj       : TObject;
                           const TableSize : TArrayIndex
                          ) : TArrayIndex; far;

{--------------------------------------------------------------------------}

type

TExtendedBucket = class(TBucket)
private
    function GetExtended : Extended;
public
    constructor Create(const Data : Extended);
    property Value : Extended read GetExtended;
end;

function CompareExtendedBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashExtendedBucket( const Obj       : TObject;
                             const TableSize : TArrayIndex
                            ) : TArrayIndex; far;

{--------------------------------------------------------------------------}

type

TCompBucket = class(TBucket)
private
    function GetComp : Comp;
public
    constructor Create(const Data : Comp);
    property Value : Comp read GetComp;
end;

function CompareCompBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashCompBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex; far;

{$ifndef Ver80}

type

TCurrencyBucket = class(TBucket)
private
    function GetCurrency : Currency;
public
    constructor Create(const Data : Currency);
    property Value : Currency read GetCurrency;
end;

function CompareCurrencyBucket(const Lhs, Rhs : TObject) : TComparison;
function HashCurrencyBucket( const Obj       : TObject;
                             const TableSize : TArrayIndex
                           ) : TArrayIndex;

{$endif}

{--------------------------------------------------------------------------}

type

TDateTimeBucket = class(TBucket)
private
    function GetDateTime : TDateTime;
public
    constructor Create(const Data : TDateTime);
    property Value : TDateTime read GetDateTime;
end;

function CompareDateTimeBucket(const Lhs, Rhs : TObject) : TComparison; far;
function HashDateTimeBucket( const Obj       : TObject;
                             const TableSize : TArrayIndex
                            ) : TArrayIndex; far;




implementation

uses
    ObjHash,  { for StringHash, FloatHash, CompHash, IntegerHash }
    SysUtils; { for NewStr }

{ TStringCombo implementation ---------------------------------------------}

constructor TStringCombo.Create(const Str : String; const Obj : TObject);
begin
    inherited Create;

    FStr       := Str;
    FObj       := Obj;
    FOwnership := True;
end;

destructor TStringCombo.Destroy;
begin
    if FOwnership then
        FObj.Free;

    inherited Destroy;
end;

function TStringCombo.GetObj    : TObject;
begin
    Result := FObj;
end;

function CompareStringCombo(const Lhs, Rhs : TObject) : TComparison;
begin
    Result := CompareString( (Lhs as TStringCombo).Str,
                             (Rhs as TStringCombo).Str );
end;

function CompareStringComboNoCase(const Lhs, Rhs : TObject) : TComparison;
begin
    Result := CompareStringNoCase( (Lhs as TStringCombo).Str,
                                   (Rhs as TStringCombo).Str );
end;

function HashStringCombo( const Obj       : TObject;
                          const TableSize : TArrayIndex
                        ) : TArrayIndex;
begin
    Result := StringHash((Obj as TStringCombo).Str, TableSize);
end;

function HashStringCombo2( const Obj       : TObject;
                          const TableSize : TArrayIndex
                        ) : TArrayIndex;
begin
    Result := StringHash2((Obj as TStringCombo).Str, TableSize);
end;

procedure ReadStringCombo( const Obj     : TObject;
                           const Reader  : TObjectReader );
var
    HasObject   : Boolean;
    StringCombo : TStringCombo;
begin
    StringCombo      := Obj as TStringCombo;
    StringCombo.FStr := Reader.ReadString;

    HasObject := Reader.ReadBoolean;

    if HasObject then
        StringCombo.FObj := Reader.ReadChildObject
    else
        StringCombo.FObj := nil;

    { Ownership is set True when reading from stream. }
    StringCombo.Ownership := True;
end;

procedure WriteStringCombo( const Obj     : TObject;
                            const Writer  : TObjectWriter );
var
    HasObject   : Boolean;
    StringCombo : TStringCombo;
begin
    StringCombo := Obj as TStringCombo;
    Writer.WriteString(StringCombo.Str);

    HasObject := (StringCombo.FObj <> nil);

    Writer.WriteBoolean(HasObject);

    if HasObject then
        Writer.WriteChildObject(StringCombo.FObj);
end;


{ TBucket implementation --------------------------------------------------}

constructor TBucket.Create(const InDataSize : Word);
begin
    inherited Create;

    FDataSize := InDataSize;

    GetMem(FDataPtr, FDataSize);

    FillChar(FDataPtr^, FDataSize, 0);
end;

{--------------------------------------------------------------------------}

constructor TBucket.CreateSetData(const InDataSize : Word; const Data);
begin
    inherited Create;

    FDataSize := InDataSize;

    GetMem(FDataPtr, FDataSize);

    Move(Data, FDataPtr^, FDataSize);
end;

{--------------------------------------------------------------------------}

destructor TBucket.Destroy;
begin
    FreeMem(FDataPtr, FDataSize);

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

procedure TBucket.SetData(const Data);
begin
    Move(Data, FDataPtr^, FDataSize);
end;
{--------------------------------------------------------------------------}

procedure TBucket.GetData(var Data);
begin
    Move(FDataPtr^, Data, FDataSize);
end;

{--------------------------------------------------------------------------}

procedure ReadBucket( const Obj     : TObject;
                      const Reader  : TObjectReader );
begin
    (Obj as TBucket).FDataSize := Reader.ReadInteger;

    GetMem((Obj as TBucket).FDataPtr, (Obj as TBucket).FDataSize);

    Reader.Read( (Obj as TBucket).FDataPtr^,
                 (Obj as TBucket).FDataSize );
end;

{--------------------------------------------------------------------------}

procedure WriteBucket( const Obj     : TObject;
                       const Writer  : TObjectWriter );
begin
    Writer.WriteInteger((Obj as TBucket).FDataSize);

    Writer.Write( (Obj as TBucket).FDataPtr^,
                  (Obj as TBucket).FDataSize );
end;


{--------------------------------------------------------------------------}

constructor TShortIntBucket.Create(const Data : ShortInt);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TShortIntBucket.GetShortInt : ShortInt;
begin
    Result := ShortInt(FDataPtr^);
end;

function CompareShortIntBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TShortIntBucket).Value) < ((Rhs as TShortIntBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TShortIntBucket).Value) > ((Rhs as TShortIntBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashShortIntBucket( const Obj       : TObject;
                             const TableSize : TArrayIndex
                            ) : TArrayIndex;
begin
    Result := IntegerHash((Obj as TShortIntBucket).Value, TableSize);
end;

{--------------------------------------------------------------------------}

constructor TIntegerBucket.Create(const Data : Integer);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TIntegerBucket.GetInteger : Integer;
begin
    Result := Integer(FDataPtr^);
end;

function CompareIntegerBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TIntegerBucket).Value) < ((Rhs as TIntegerBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TIntegerBucket).Value) > ((Rhs as TIntegerBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashIntegerBucket( const Obj       : TObject;
                            const TableSize : TArrayIndex
                           ) : TArrayIndex;
begin
    Result := IntegerHash((Obj as TIntegerBucket).Value, TableSize);
end;

{--------------------------------------------------------------------------}

constructor TLongIntBucket.Create(const Data : LongInt);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TLongIntBucket.GetLongInt : LongInt;
begin
    Result := LongInt(FDataPtr^);
end;

function CompareLongIntBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TLongIntBucket).Value) < ((Rhs as TLongIntBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TLongIntBucket).Value) > ((Rhs as TLongIntBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashLongIntBucket( const Obj       : TObject;
                            const TableSize : TArrayIndex
                           ) : TArrayIndex;
begin
    Result := IntegerHash((Obj as TLongIntBucket).Value, TableSize);
end;


{--------------------------------------------------------------------------}

constructor TByteBucket.Create(const Data : Byte);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TByteBucket.GetByte : Byte;
begin
    Result := Byte(FDataPtr^);
end;

function CompareByteBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TByteBucket).Value) < ((Rhs as TByteBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TByteBucket).Value) > ((Rhs as TByteBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashByteBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex;
begin
    Result := IntegerHash((Obj as TByteBucket).Value, TableSize); 
end;

{--------------------------------------------------------------------------}

constructor TWordBucket.Create(const Data : Word);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TWordBucket.GetWord : Word;
begin
    Result := Word(FDataPtr^);
end;

function CompareWordBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TWordBucket).Value) < ((Rhs as TWordBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TWordBucket).Value) > ((Rhs as TWordBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashWordBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex;
begin
    Result := IntegerHash((Obj as TWordBucket).Value, TableSize)
end;

{--------------------------------------------------------------------------}

constructor TCharBucket.Create(const Data : Char);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TCharBucket.GetChar : Char;
begin
    Result := Char(FDataPtr^);
end;

function CompareCharBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TCharBucket).Value) < ((Rhs as TCharBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TCharBucket).Value) > ((Rhs as TCharBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashCharBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex;
begin
    Result := IntegerHash(Ord((Obj as TCharBucket).Value), TableSize);
end;

{--------------------------------------------------------------------------}

constructor TBooleanBucket.Create(const Data : Boolean);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TBooleanBucket.GetBoolean : Boolean;
begin
    Result := Boolean(FDataPtr^);
end;

function CompareBooleanBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TBooleanBucket).Value) < ((Rhs as TBooleanBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TBooleanBucket).Value) > ((Rhs as TBooleanBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

{--------------------------------------------------------------------------}

constructor TWordBoolBucket.Create(const Data : WordBool);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TWordBoolBucket.GetWordBool : WordBool;
begin
    Result := WordBool(FDataPtr^);
end;

function CompareWordBoolBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TWordBoolBucket).Value) < ((Rhs as TWordBoolBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TWordBoolBucket).Value) > ((Rhs as TWordBoolBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

{--------------------------------------------------------------------------}

constructor TLongBoolBucket.Create(const Data : LongBool);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TLongBoolBucket.GetLongBool : LongBool;
begin
    Result := LongBool(FDataPtr^);
end;

function CompareLongBoolBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TLongBoolBucket).Value) < ((Rhs as TLongBoolBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TLongBoolBucket).Value) > ((Rhs as TLongBoolBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

{--------------------------------------------------------------------------}

constructor TRealBucket.Create(const Data : Real);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TRealBucket.GetReal : Real;
begin
    Result := Real(FDataPtr^);
end;

function CompareRealBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TRealBucket).Value) < ((Rhs as TRealBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TRealBucket).Value) > ((Rhs as TRealBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashRealBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex;
begin
    Result := FloatHash((Obj as TRealBucket).Value, TableSize);
end;

{--------------------------------------------------------------------------}

constructor TSingleBucket.Create(const Data : Single);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TSingleBucket.GetSingle : Single;
begin
    Result := Single(FDataPtr^);
end;

function CompareSingleBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TSingleBucket).Value) < ((Rhs as TSingleBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TSingleBucket).Value) > ((Rhs as TSingleBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashSingleBucket( const Obj       : TObject;
                           const TableSize : TArrayIndex
                          ) : TArrayIndex;
begin
    Result := FloatHash((Obj as TSingleBucket).Value, TableSize);
end;

{--------------------------------------------------------------------------}

constructor TDoubleBucket.Create(const Data : Double);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TDoubleBucket.GetDouble : Double;
begin
    Result := Double(FDataPtr^);
end;

function CompareDoubleBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TDoubleBucket).Value) < ((Rhs as TDoubleBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TDoubleBucket).Value) > ((Rhs as TDoubleBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashDoubleBucket( const Obj       : TObject;
                           const TableSize : TArrayIndex
                          ) : TArrayIndex;
begin
    Result := FloatHash((Obj as TDoubleBucket).Value, TableSize);
end;

{--------------------------------------------------------------------------}

constructor TExtendedBucket.Create(const Data : Extended);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TExtendedBucket.GetExtended : Extended;
begin
    Result := Extended(FDataPtr^);
end;

function CompareExtendedBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TExtendedBucket).Value) < ((Rhs as TExtendedBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TExtendedBucket).Value) > ((Rhs as TExtendedBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashExtendedBucket( const Obj       : TObject;
                             const TableSize : TArrayIndex
                            ) : TArrayIndex;
begin
    Result := FloatHash((Obj as TExtendedBucket).Value, TableSize);
end;

{--------------------------------------------------------------------------}

constructor TCompBucket.Create(const Data : Comp);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TCompBucket.GetComp : Comp;
begin
    Result := Comp(FDataPtr^);
end;

function CompareCompBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TCompBucket).Value) < ((Rhs as TCompBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TCompBucket).Value) > ((Rhs as TCompBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashCompBucket( const Obj       : TObject;
                         const TableSize : TArrayIndex
                        ) : TArrayIndex;
begin
    Result := FloatHash((Obj as TCompBucket).Value, TableSize);
end;

{--------------------------------------------------------------------------}

{$ifndef Ver80}

constructor TCurrencyBucket.Create(const Data : Currency);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;
                  
function TCurrencyBucket.GetCurrency : Currency;
begin
    Result := Currency(FDataPtr^);
end;

function CompareCurrencyBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TCurrencyBucket).Value) < ((Rhs as TCurrencyBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TCurrencyBucket).Value) > ((Rhs as TCurrencyBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashCurrencyBucket( const Obj       : TObject;
                             const TableSize : TArrayIndex
                           ) : TArrayIndex;
begin
    Result := FloatHash((Obj as TCurrencyBucket).Value, TableSize);
end;

{$endif}

{--------------------------------------------------------------------------}

constructor TDateTimeBucket.Create(const Data : TDateTime);
begin
    inherited CreateSetData(SizeOf(Data), Data);
end;

function TDateTimeBucket.GetDateTime : TDateTime;
begin
    Result := TDateTime(FDataPtr^);
end;

function CompareDateTimeBucket(const Lhs, Rhs : TObject) : TComparison;
begin
    if ((Lhs as TDateTimeBucket).Value) < ((Rhs as TDateTimeBucket).Value) then
        Result := cmpLess
    else if ((Lhs as TDateTimeBucket).Value) > ((Rhs as TDateTimeBucket).Value) then
        Result := cmpGreater
    else
        Result := cmpEqual;
end;

function HashDateTimeBucket( const Obj       : TObject;
                             const TableSize : TArrayIndex
                           ) : TArrayIndex;
begin
    Result := FloatHash((Obj as TDateTimeBucket).Value, TableSize);
end;

{--------------------------------------------------------------------------}

initialization

    {$ifdef Win32}

    RegisterStreamable( TStringCombo,
                        [ TObjectReadProc(@ReadStringCombo) ],
                        [ TObjectWriteProc(@WriteStringCombo) ]
                      );

    RegisterStreamable( TBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TShortIntBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TIntegerBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TLongIntBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TByteBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TWordBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TCharBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TBooleanBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TWordBoolBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TLongBoolBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TRealBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TDateTimeBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TSingleBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TDoubleBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TExtendedBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );

    RegisterStreamable( TCompBucket,
                        [ TObjectReadProc(@ReadBucket) ],
                        [ TObjectWriteProc(@WriteBucket) ]
                       );
    
    {$else}

    RegisterStreamable( TStringCombo,
                        [ReadStringCombo],
                        [WriteStringCombo]
                      );

    RegisterStreamable( TBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TShortIntBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TIntegerBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TLongIntBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TByteBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TWordBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TCharBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TBooleanBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TWordBoolBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TLongBoolBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TRealBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TDateTimeBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TSingleBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TDoubleBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TExtendedBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    RegisterStreamable( TCompBucket,
                        [ReadBucket],
                        [WriteBucket]
                       );

    {$endif}


    RegisterCompareFunction( 'CompareStringCombo', CompareStringCombo );
    RegisterCompareFunction( 'CompareStringComboNoCase', CompareStringComboNoCase );
    RegisterCompareFunction( 'CompareShortIntBucket', CompareShortIntBucket );
    RegisterCompareFunction( 'CompareIntegerBucket', CompareIntegerBucket );
    RegisterCompareFunction( 'CompareLongIntBucket', CompareLongIntBucket );
    RegisterCompareFunction( 'CompareByteBucket', CompareByteBucket );
    RegisterCompareFunction( 'CompareWordBucket', CompareWordBucket );
    RegisterCompareFunction( 'CompareCharBucket', CompareCharBucket );
    RegisterCompareFunction( 'CompareBooleanBucket', CompareBooleanBucket );
    RegisterCompareFunction( 'CompareWordBoolBucket', CompareWordBoolBucket );
    RegisterCompareFunction( 'CompareLongBoolBucket', CompareLongBoolBucket );
    RegisterCompareFunction( 'CompareRealBucket', CompareRealBucket );
    RegisterCompareFunction( 'CompareDateTimeBucket', CompareDateTimeBucket );

    RegisterCompareFunction( 'CompareSingleBucket', CompareSingleBucket );
    RegisterCompareFunction( 'CompareDoubleBucket', CompareDoubleBucket );
    RegisterCompareFunction( 'CompareExtendedBucket', CompareExtendedBucket );
    RegisterCompareFunction( 'CompareCompBucket', CompareCompBucket );

    RegisterHashFunction( 'HashStringCombo', HashStringCombo );
    RegisterHashFunction( 'HashStringCombo2', HashStringCombo2 );
    RegisterHashFunction( 'HashShortIntBucket', HashShortIntBucket );
    RegisterHashFunction( 'HashIntegerBucket', HashIntegerBucket );
    RegisterHashFunction( 'HashLongIntBucket', HashLongIntBucket );
    RegisterHashFunction( 'HashByteBucket', HashByteBucket );
    RegisterHashFunction( 'HashWordBucket', HashWordBucket );
    RegisterHashFunction( 'HashCharBucket', HashCharBucket );
    RegisterHashFunction( 'HashRealBucket', HashRealBucket );
    RegisterHashFunction( 'HashDateTimeBucket', HashDateTimeBucket );

    RegisterHashFunction( 'HashSingleBucket', HashSingleBucket );
    RegisterHashFunction( 'HashDoubleBucket', HashDoubleBucket );
    RegisterHashFunction( 'HashExtendedBucket', HashExtendedBucket );
    RegisterHashFunction( 'HashCompBucket', HashCompBucket );
end.
