{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjHash.Pas
|
| Description: Hash table container.
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| History: Version 1.0   Michel Brazeau, March 1996.
|          Oct 31, 1996. Michel Brazeau, Improved StringHash function.
|          Dec 11, 1996. Michel Brazeau, Added Detach method.
|          May 11, 1997. Michel Brazeau, move THashFunc type
|                                     to ObjCmp unit to prevent circular
|                                     unit references.
|          Jul 07, 1998. MB add StringHash2
|          Jul 09, 1998. MB add FloatHast,
|
|---------------------------------------------------------------------------}
unit ObjHash;

{$I Contain.Inc}

interface

uses
    ObjPerst, { for TObjectReader, TObjectWriter }
    ObjCmp,   { for TCompareFunc }
    ObjType,  { for for TArrayIndex }
    ObjArray, { for TArray, TArrayIndex }
    ObjBase;  { for TContainer }


{ Returns a hash index given an integer value. Uses the modulus operator }
function IntegerHash( const Value : LongInt;
                      const TableSize : TArrayIndex) : TArrayIndex; far;

{ Returns a hash index given a string. Hash is based on the first 5
  characters of the string. }
function StringHash( const Value     : String;
                     const TableSize : TArrayIndex) : TArrayIndex; far;

{ Returns a hash index given a string. Hash is based on the first 5
  characters of the string. }
function StringHash2( const Value     : String;
                      const TableSize : TArrayIndex) : TArrayIndex; far;


{ Returns a hash index given a floating point value. }
function FloatHash( const Value     : Extended;
                    const TableSize : TArrayIndex) : TArrayIndex; far;


type

THashTable = class(TContainer)
public
    constructor Create( const InObjectType : TClass;
                        const InCompareFunc : TCompareFunc;
                        const InHashFunc    : THashFunc;
                        const InTableSize   : TArrayIndex );

    destructor Destroy; override;

    procedure Clear; override;

    procedure Insert(const Obj : TObject); virtual;

    { Deletes (removes and Frees) the N'th matching Occur'ence of Obj.
      Returns True if the object is found and deleted. }
    function Delete( const Obj   : TObject;
                     const Occur : LongInt
                    ) : Boolean; virtual;


    { Detaches and returns the N'th matching Occur'ence of Obj.
      Returns nil if an object matching Obj is not found. }
    function Detach( const Obj   : TObject;
                     const Occur : LongInt
                    ) : TObject;

    { searched for the N'th occurence of Obj. Returns the reference to
      the found object if found. If not found, nil is returned }
    function Search( const Obj    : TObject;    { key to search }
                     const Occur  : LongInt     { occurence }
                    ) : TObject;

    procedure ForEachCallProc( const ProcIterator : TProcIterator;
                               const CallBackData : LongInt); override;

    procedure ForEachCallMethod( const MethodIterator : TMethodIterator;
                                 const CallBackData   : LongInt); override;

protected
    HashArray  : TArray;      { array of link lists for each table entry }

    FHashFunc  : THashFunc;   { hash function }

    FTableSize : TArrayIndex; { size of hash table }

    procedure SetOwnership( const InOwnership : Boolean ); override;
end;

{-------------------------------------------------------------------------}

{ Streaming routines }

procedure ReadHashTable( const Obj    : TObject;
                         const Reader : TObjectReader ); far;

procedure WriteHashTable( const Obj    : TObject;
                          const Writer : TObjectWriter ); far;

implementation

uses
    ObjBuckt, { for TStringCombo }
    ObjList;  { for TUnOrderedList }

{--------------------------------------------------------------------------}

function IntegerHash(const Value : LongInt; const TableSize : TArrayIndex) : TArrayIndex;
begin
    { MB Oct 22, 2001. Ensure negative value is not returned using Abs }
    Result := Abs(Value) mod TableSize + 1;
end;

{--------------------------------------------------------------------------}

function StringHash(const Value : String; const TableSize : TArrayIndex) : TArrayIndex;
const
    CHashConst = 23; { prime number. If changed, ensure it does not create
                       an overflow in the Sum variable. }
var
    I     : Word;
    Max   : Word;
    Sum   : LongInt;

begin
    Max   := Length(Value);
    if Max > 5 then  { Hash on the first 5 characters }
        Max := 5;

    I   := 1;
    Sum := 0;

    while (I <= Max) do
    begin
        Sum := (Sum * CHashConst) + Ord(Value[I]);
        Inc(I)
    end;

    { keep hash address in range }
    Result := (Sum mod TableSize) + 1;
end;

{--------------------------------------------------------------------------}

function StringHash2(const Value : String; const TableSize : TArrayIndex) : TArrayIndex;
const
    CHashConst   = 13;
    CSampleCount =  7;
var
    Position  : Integer;
    StrLength : Integer;
    Sum       : LongInt;
    Delta     : Integer;
    Count     : Integer;

begin
    StrLength := Length(Value);
    Delta     := StrLength div CSampleCount;
    if Delta = 0 then
        Delta := 1;

    Position  := 1;
    Sum       := 0;
    Count     := 1;

    while (Position <= StrLength) and
          (Count    <=  CSampleCount) do
    begin
        if Count = CSampleCount then
            Position := StrLength;

        Sum := (Sum * CHashConst) + Ord(Value[Position]);
        Inc(Position, Delta);
        Inc(Count);
    end;

    { keep hash address in range }
    Result := (Sum mod TableSize) + 1;
end;

{--------------------------------------------------------------------------}

function FloatHash( const Value     : Extended;
                    const TableSize : TArrayIndex) : TArrayIndex;
var
    Division  : Extended;
    Remainder : LongInt;
begin
    Division  := (Value / TableSize);
    Remainder := Round(Value - (Division * TableSize));
    Result := Remainder + 1;
end;

{--------------------------------------------------------------------------}

constructor THashTable.Create( const InObjectType  : TClass;
                               const InCompareFunc : TCompareFunc;
                               const InHashFunc    : THashFunc;
                               const InTableSize   : TArrayIndex );
var
    I : TArrayIndex;
begin
    FHashFunc  := InHashFunc;

    if InTableSize < 1 then
        RaiseInvalidArrayCapacity;

    FTableSize := InTableSize;

    HashArray  := TArray.Create(TUnOrderedList, CompareEqual, FTableSize, 0);

    { insert a list for every hash entry }
    for I := 1 to FTableSize do
    begin
        HashArray.Insert(TUnOrderedList.Create(InObjectType, InCompareFunc));
    end;

    inherited Create(InObjectType);

    { MB May 12, 1997. Set the base class FCompareFunc for streaming purposes }
    FCompareFunc := InCompareFunc;

end;

{--------------------------------------------------------------------------}

destructor THashTable.Destroy;
begin
    { must call the base destructor first so that HashTable is still valid
      when the call to Clear is made in the base class destructor }
    inherited Destroy;

    HashArray.Free;
end;

{--------------------------------------------------------------------------}

procedure THashTable.SetOwnership( const InOwnership : Boolean );
var
    I : TArrayIndex;
begin
    { maintain the FOwnership data member in the base class }
    inherited SetOwnership(InOwnership);

    for I := 1 to FTableSize do
        (HashArray[I] as TUnOrderedList).Ownership := InOwnership;
end;

{--------------------------------------------------------------------------}

procedure THashTable.ForEachCallProc( const ProcIterator : TProcIterator;
                                      const CallBackData : LongInt);
var
    I : TArrayIndex;
begin
    for I := 1 to FTableSize do
        (HashArray[I] as TUnOrderedList).ForEachCallProc(ProcIterator, CallBackData);
end;

{--------------------------------------------------------------------------}

procedure THashTable.ForEachCallMethod( const MethodIterator : TMethodIterator;
                                        const CallBackData   : LongInt );
var
    I : TArrayIndex;
begin
    for I := 1 to FTableSize do
        (HashArray[I] as TUnOrderedList).ForEachCallMethod(MethodIterator, CallBackData);
end;

{--------------------------------------------------------------------------}

procedure THashTable.Clear;
var
    I : TArrayIndex;
begin
    for I := 1 to FTableSize do
    begin
        { check for nil, as Clear may be invoked on a partially constructed
          object }
        if HashArray <> nil then
            (HashArray[I] as TUnOrderedList).Clear;
    end;

    FSize := 0;
end;

{--------------------------------------------------------------------------}

procedure THashTable.Insert(const Obj : TObject);
var
    List : TUnOrderedList;
begin
    List := HashArray[FHashFunc(Obj, FTableSize)] as TUnOrderedList;

    { add the object to the proper hash entry list }
    List.InsertAtTail(Obj);

    Inc(FSize);
end;

{--------------------------------------------------------------------------}

function THashTable.Delete( const Obj   : TObject;
                            const Occur : LongInt
                           ) : Boolean;
var
    List : TUnOrderedList;
begin
    List   := HashArray[FHashFunc(Obj, FTableSize)] as TUnOrderedList;

    Result := List.Delete(Obj, Occur);

    if Result then
        Dec(FSize);
end;

{--------------------------------------------------------------------------}

function THashTable.Detach( const Obj   : TObject;
                            const Occur : LongInt
                           ) : TObject;
var
    List : TUnOrderedList;
begin
    List   := HashArray[FHashFunc(Obj, FTableSize)] as TUnOrderedList;

    if List.Search(Obj, Occur) then
    begin
        Result := List.DetachCurrent;
        Dec(FSize);
    end
    else
        Result := nil;
end;

{--------------------------------------------------------------------------}

function THashTable.Search( const Obj   : TObject;
                            const Occur : LongInt
                           ) : TObject;
var
    List : TUnOrderedList;
begin
    List   := HashArray[FHashFunc(Obj, FTableSize)] as TUnOrderedList;

    if List.Search(Obj, Occur) then
        Result := List.CurrentObj
    else
        Result := nil;
end;

{-------------------------------------------------------------------------}

{ Streaming routines }

procedure ReadHashTable( const Obj    : TObject;
                         const Reader : TObjectReader );
var
    HashTable : THashTable;
    Size      : LongInt;
    Index     : LongInt;
    Item      : TObject;

begin
    HashTable := Obj as THashTable;

    ReadContainer(Obj, Reader);

    Size := Reader.ReadInteger;

    HashTable.FHashFunc  := GetHashFunc(Reader.ReadString);

    HashTable.FTableSize := Reader.ReadInteger;

    HashTable.HashArray  := TArray.Create( TUnOrderedList, CompareEqual,
                                           HashTable.FTableSize, 0);

    { insert a list for every hash entry }
    for Index := 1 to HashTable.FTableSize do
    begin
        HashTable.HashArray.Insert(TUnOrderedList.Create( HashTable.FObjectType,
                                                HashTable.FCompareFunc));
    end;

    for Index := 1 to Size do
    begin
        Item := Reader.ReadChildObject;

        HashTable.Insert(Item);
    end;
end;

{-------------------------------------------------------------------------}


procedure StreamObject( const Obj          : TObject;
                        const CallBackData : LongInt ); far;
begin
    TObjectWriter(CallBackData).WriteChildObject(Obj);
end;

procedure WriteHashTable( const Obj    : TObject;
                          const Writer : TObjectWriter );
var
    HashTable : THashTable;

begin
    HashTable := Obj as THashTable;

    WriteContainer(HashTable, Writer);

    Writer.WriteInteger(HashTable.Size);

    Writer.WriteString(GetHashFuncName(HashTable.FHashFunc));

    Writer.WriteInteger(HashTable.FTableSize);

    HashTable.ForEachCallProc(StreamObject, LongInt(Writer));

end;

{-------------------------------------------------------------------------}

initialization

    {$ifdef Win32}
    RegisterStreamable( THashTable,
                        [TObjectReadProc(@ReadHashTable)],
                        [TObjectWriteProc(@WriteHashTable)] );
    {$else}
    RegisterStreamable( THashTable,
                        [ReadHashTable],
                        [WriteHashTable] );
    {$endif}

end.
