{----------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: DriverTest.Pas
|
| Description: Test cases used to test TOdbDriver drived classes.
|
| History: Feb 16, 2001. Michel Brazeau, first version
|
|---------------------------------------------------------------------}
unit DriverTest;

{$I Odb.Inc}

{$ifndef __Testing}
    !! This unit is for automated tests only
{$endif}

interface

uses
    AutoTest, { for TSpiderOdbTestcase }
    Classes,  { for TStringList }
    TestObj,  { for TTestObject }
    Odb;      { for TOdbDriverClass }

{--------------------------------------------------------------------------}

type

TOdbDriverTest = class( TSpiderOdbTestCase )
protected
    FObjectDatabase    : TObjectDatabase;
    FObjectFile        : TObjectFile;
    FObjectQuery       : TObjectQuery;
    FObjectUpdate      : TObjectUpdate;
    FDriverClass       : TOdbDriverClass;

    FDatabaseName      : String;
    FServerName        : String;
    FUserName          : String;
    FPassword          : String;

    FDriverName        : String;

    FFirstAutoIncValue : LongInt;

    procedure GetParams( var DriverClass                   : TOdbDriverClass;
                         var DatabaseName                  : String;
                         var ServerName                    : String;
                         var UserName                      : String;
                         var Password                      : String
                       ); virtual; abstract;

    procedure Setup; override;
    procedure Teardown; override;



    { Traversal tests methods on specific indices }
    procedure TestTraverseByNameISAM;
    procedure TestTraverseByIdISAM;
    procedure TestTraverseByAgeISAM;
    procedure TestTraverseByWeightISAM;

    procedure TestFindOnNameISAM;
    procedure TestFindOnIdISAM;
    procedure TestFindOnAgeISAM;
    procedure TestFindOnWeightISAM;
    procedure TestFindOnNameCiISAM; { case insensitive }
    procedure TestFindOnAgeNameISAM;
    procedure TestFindOnAgeDescISAM; { descending index }
    procedure TestFindOnWeightNameISAM;
    procedure TestFindOnWeightAgeISAM;

    procedure TestTraverseByNameSQL;
    procedure TestTraverseByAgeSQL;
    procedure TestTraverseByWeightSQL;

    procedure TestFindOnNameSQL;
    procedure TestFindOnAgeSQL;
    procedure TestFindOnWeightSQL;

    procedure TestUpdateISAM;
    procedure TestDuplicateInsertISAM;
    procedure TestDuplicateUpdateISAM;
    procedure TestDeleteObjectFile;

    procedure TestInvalidIndexName;

    procedure TestTransactionsSQL;

    procedure TestUpdateSQL;
    procedure TestDuplicateInsertSQL;
    procedure TestDuplicateUpdateSQL;
    procedure TestDeleteObjectQuery;

    procedure TestObjectFileLockingAndPack;

    procedure TestTransactionsISAM;

    procedure TestObjectLockingObjectFile;

    procedure TestCreateDeleteFile;

    procedure TestBookmarks;


    procedure TestObjectFileHooks( const ObjectFile : TObjectFile );


published


    procedure TestISAM;

    procedure TestSQL;


end;


(************************************************************************

  * *   * ***  *    **** *   * **** *   * *****   *   ***** *  **  *   *
  * ** ** *  * *    *    ** ** *    **  *   *    * *    *   * *  * **  *
  * * * * ***  *    **   * * * **   * * *   *   *   *   *   * *  * * * *
  * *   * *    *    *    *   * *    *  **   *   *****   *   * *  * *  **
  * *   * *    **** **** *   * **** *   *   *   *   *   *   *  **  *   *

 ************************************************************************)

implementation

uses
    OdbBde,   { for TBdeOdbDriver }
    SysUtils; { for IntToStr }

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.Setup;
begin

    GetParams( FDriverClass, FDatabaseName, FServerName, FUserName, FPassword );

    FDriverName := FDriverClass.ClassName;

    FObjectDatabase := TObjectDatabase.Create( FDriverClass, FDatabaseName );

    FObjectDatabase.Open( FServerName, FUserName, FPassword );

    if damISAM in FObjectDatabase.Driver.GetDataAccessMethods then
        FObjectFile  := TTestObjectFile.Create( FObjectDatabase, 'TestData', TTestObjectMapping  );

    if damSQL in FObjectDatabase.Driver.GetDataAccessMethods then
    begin
        FObjectQuery  := TObjectQuery.Create( FObjectDatabase, 'TestData', TTestObjectMapping );
        FObjectUpdate := TObjectUpdate.Create( FObjectDatabase, 'TestData', TTestObjectMapping );
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.Teardown;
begin
    FObjectUpdate.Free;
    FObjectQuery.Free;
    FObjectFile.Free;

    FObjectDatabase.Close;

    FObjectDatabase.Free;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestISAM;
var
    TestObject : TTestObject;
    Raised     : Boolean;
begin
    if not (damISAM in FObjectDatabase.Driver.GetDataAccessMethods) then
        Exit;

    Check( FObjectFile.Database = FObjectDatabase );

    Check( AnsiUppercase(FObjectFile.DatasetName)    = AnsiUppercase('TestData'));
    Check( AnsiUppercase(FObjectFile.ObjectFileName) = AnsiUppercase('TestData'));

    Raised := False;
    try
        Check( FObjectFile.FieldByName('UnknownFld') = nil);
    except
        on E:EOdbFieldDoesNotExist do
            Raised := True
        else
            raise;
    end;
    if not Raised then
        Check(False, 'FieldByName failed');

    Check( FObjectFile.Driver is FDriverClass);

    FObjectFile.CreateDataset;
    TestObject := TTestObject.Create;
    try
        Check( not FObjectFile.Opened );

        TestObjectFileHooks(FObjectFile);

        FObjectFile.Open;
        try
            Check( FObjectFile.IsEmpty );

            Check( FObjectFile.Opened );

            AddObjectISAM( 1, FObjectFile );

            { DBISAM does not start its first auto inc at 1, so the initial value needs to be recorded }
            FFirstAutoIncValue := (FObjectFile.CurrentObj as TTestObject).Id;

            AddObjectISAM( 2, FObjectFile );
            AddObjectISAM( 3, FObjectFile );
            AddObjectISAM( 4, FObjectFile );
            AddObjectISAM( 5, FObjectFile );
            AddObjectISAM( 6, FObjectFile );
            AddObjectISAM( 7, FObjectFile );
            AddObjectISAM( 8, FObjectFile );
            AddObjectISAM( 9, FObjectFile );

            if FObjectDatabase.Driver.AutoIncSupported then
                Check( (FObjectFile.CurrentObj as TTestObject).Id = FFirstAutoIncValue + 8, 'AutoInc failed');

            Check( FObjectFile.ObjectCount =  9, 'ObjectCount <> 9');

            TestInvalidIndexName;

            TestTraverseByNameISAM;
            TestTraverseByIdISAM;
            TestTraverseByAgeISAM;
            TestTraverseByWeightISAM;

            TestFindOnNameISAM;
            TestFindOnIdISAM;
            TestFindOnAgeISAM;
            TestFindOnWeightISAM;
            TestFindOnNameCiISAM;
            TestFindOnAgeNameISAM;
            TestFindOnAgeDescISAM;
            TestFindOnWeightNameISAM;
            TestFindOnWeightAgeISAM;

            TestBookmarks;

            TestTransactionsISAM;

            TestObjectLockingObjectFile;

            TestUpdateISAM;

            TestDuplicateInsertISAM;

            { this causes TestDeleteObjectFile to fail with FlashFiler }
            TestDuplicateUpdateISAM;

            TestDeleteObjectFile;

            TestObjectFileLockingAndPack;

        finally
            FObjectFile.Close;

            Check( not FObjectFile.Opened );
        end;

    finally
        TestObject.Free;
        FObjectFile.DeleteDataset;
    end;


    TestCreateDeleteFile;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestInvalidIndexName;
begin
    try
        FObjectFile.IndexName := 'Customer';
    except
        Exit;
    end;
    Check( False, 'TestInvalidIndexName');
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTraverseByNameISAM;
var
    Count : Integer;
begin
    FObjectFile.IndexName := 'Name';

    Check(FObjectFile.CurrentIndex.Name = 'Name');

    { check forward traversal }
    Check( FObjectFile.FindFirst );
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    for Count := 2 to 9 do
    begin
        Check( FObjectFile.FindNext );
        CheckObjectValues( Count, FObjectFile.CurrentObj as TTestObject );
    end;
    Check( not FObjectFile.FindNext );

    { check backward traversal }
    Check( FObjectFile.FindLast );
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    for Count := 8 downto 1 do
    begin
        Check( FObjectFile.FindPrior );
        CheckObjectValues( Count, FObjectFile.CurrentObj as TTestObject );
    end;
    Check( not FObjectFile.FindPrior );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTraverseByIdISAM;
var
    Count : Integer;
begin
    if not FObjectDatabase.Driver.AutoIncSupported then
        Exit;

    FObjectFile.IndexName := 'Id';

    Check(FObjectFile.CurrentIndex.Name = 'Id');

    { check forward traversal }
    Check( FObjectFile.FindFirst );
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    for Count := 2 to 9 do
    begin
        Check( FObjectFile.FindNext );
        CheckObjectValues( Count, FObjectFile.CurrentObj as TTestObject );
    end;
    Check( not FObjectFile.FindNext );

    { check backward traversal }
    Check( FObjectFile.FindLast );
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    for Count := 8 downto 1 do
    begin
        Check( FObjectFile.FindPrior );
        CheckObjectValues( Count, FObjectFile.CurrentObj as TTestObject );
    end;
    Check( not FObjectFile.FindPrior );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTraverseByAgeISAM;
var
    Count : Integer;
begin
    if not FObjectDatabase.Driver.IntegerIndexSupported then
        Exit;

    FObjectFile.IndexName := 'Age';

    Check(FObjectFile.CurrentIndex.Name = 'Age');

    { check forward traversal }
    Check( FObjectFile.FindFirst );
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    for Count := 2 to 9 do
    begin
        Check( FObjectFile.FindNext );
        CheckObjectValues( Count, FObjectFile.CurrentObj as TTestObject );
    end;
    Check( not FObjectFile.FindNext );

    { check backward traversal }
    Check( FObjectFile.FindLast );
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    for Count := 8 downto 1 do
    begin
        Check( FObjectFile.FindPrior );
        CheckObjectValues( Count, FObjectFile.CurrentObj as TTestObject );
    end;
    Check( not FObjectFile.FindPrior );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTraverseByWeightISAM;
var
    Count : Integer;
begin
    if not FObjectDatabase.Driver.FloatIndexSupported then
        Exit;

    FObjectFile.IndexName := 'Weight';

    Check(FObjectFile.CurrentIndex.Name = 'Weight');

    { check forward traversal. The weight value decreases as the other
      fields inrease. For example, the weight for 'Object1' is larger than
      the weight for 'Object2'. }
    Check( FObjectFile.FindFirst );
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    for Count := 2 to 9 do
    begin
        Check( FObjectFile.FindNext );
        CheckObjectValues( 10-Count, FObjectFile.CurrentObj as TTestObject );
    end;
    Check( not FObjectFile.FindNext );

    { check backward traversal }
    Check( FObjectFile.FindLast );
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    for Count := 8 downto 1 do
    begin
        Check( FObjectFile.FindPrior );
        CheckObjectValues( 10-Count, FObjectFile.CurrentObj as TTestObject );
    end;
    Check( not FObjectFile.FindPrior );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnNameISAM;
begin
    FObjectFile.IndexName := 'Name';

    (FObjectFile.CurrentObj as TTestObject).Name := 'Objct1';
    Check( not FObjectFile.FindExact );

    (FObjectFile.CurrentObj as TTestObject).Name := 'OBjecT1';
    Check( not FObjectFile.FindExact );

    (FObjectFile.CurrentObj as TTestObject).Name := 'Object5';
    Check( FObjectFile.FindExact );
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Name := 'Object35';
    FObjectFile.FindNearest;
    CheckObjectValues( 4, FObjectFile.CurrentObj as TTestObject );

    { while testing with QDB this ensured the BOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindLast;

    (FObjectFile.CurrentObj as TTestObject).Name := 'Object0';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );

    (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );


    { while testing with QDB this ensured the EOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindFirst;

    (FObjectFile.CurrentObj as TTestObject).Name := 'Object9';
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    { finding beyond the last object must position the last object }
    FObjectFile.FindFirst;
    (FObjectFile.CurrentObj as TTestObject).Name := 'ObjectZ';
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnIdISAM;
begin
    if not FObjectDatabase.Driver.AutoIncSupported then
        Exit;

    FObjectFile.IndexName := 'Id';

    (FObjectFile.CurrentObj as TTestObject).Id := FFirstAutoIncValue + 123;
    Check( not FObjectFile.FindExact );

    (FObjectFile.CurrentObj as TTestObject).Id := FFirstAutoIncValue - 1 + 5;
    Check( FObjectFile.FindExact );
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Id := 0;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );

    { while testing with QDB this ensured the BOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindLast;

    (FObjectFile.CurrentObj as TTestObject).Id := 0;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );


    (FObjectFile.CurrentObj as TTestObject).Id := FFirstAutoIncValue;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );


    { while testing with QDB this ensured the EOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindFirst;

    (FObjectFile.CurrentObj as TTestObject).Id := FFirstAutoIncValue - 1 + 9;
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    { finding beyond the last object must position the last object }
    (FObjectFile.CurrentObj as TTestObject).Id := 987733;
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnAgeISAM;
begin
    if not FObjectDatabase.Driver.IntegerIndexSupported then
        Exit;

    FObjectFile.IndexName := 'Age';

    (FObjectFile.CurrentObj as TTestObject).Age := 106;
    Check( not FObjectFile.FindExact );

    (FObjectFile.CurrentObj as TTestObject).Age := 5;
    Check( FObjectFile.FindExact );
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Age := 0;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );

    { while testing with QDB this ensured the BOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindLast;

    (FObjectFile.CurrentObj as TTestObject).Age := 0;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );


    (FObjectFile.CurrentObj as TTestObject).Age := 1;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );


    { while testing with QDB this ensured the EOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindFirst;

    (FObjectFile.CurrentObj as TTestObject).Age := 9;
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    { finding beyond the last object must position the last object }
    (FObjectFile.CurrentObj as TTestObject).Age := 99883;
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnWeightISAM;
begin
    if not FObjectDatabase.Driver.FloatIndexSupported then
        Exit;

    FObjectFile.IndexName := 'Weight';

    { float values
    Object9 = -121.950
    Object8 = -108.400
    Object7 = -94.850
    Object6 = -81.300
    Object5 = -67.750
    Object4 = -54.200
    Object3 = -40.650
    Object2 = -27.100
    Object1 = -13.550
    }

    (FObjectFile.CurrentObj as TTestObject).Weight := 106.33;
    Check( not FObjectFile.FindExact );

    { don't try a FindExact on float fields, just try a FindNearest }
    (FObjectFile.CurrentObj as TTestObject).Weight := -67.7511;
    FObjectFile.FindNearest;
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Weight := -150;
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );

    { while testing with QDB this ensured the BOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindLast;

    (FObjectFile.CurrentObj as TTestObject).Weight := -121.96;
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );

    { while testing with QDB this ensured the EOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindFirst;

    { finding beyond the last object must position the last object }
    (FObjectFile.CurrentObj as TTestObject).Weight := -13.6;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnNameCiISAM;
begin
    if not FObjectDatabase.Driver.CaseInsensitiveIndexSupported or
       FObjectDatabase.Driver.SingleIndexOnly then
        Exit;

    FObjectFile.IndexName := 'NameCI';

    (FObjectFile.CurrentObj as TTestObject).Name := 'Objct1';
    Check( not FObjectFile.FindExact );

    (FObjectFile.CurrentObj as TTestObject).Name := 'OBjeCt5';
    Check( FObjectFile.FindExact );
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Name := 'ObjEct35';
    FObjectFile.FindNearest;
    CheckObjectValues( 4, FObjectFile.CurrentObj as TTestObject );

    { while testing with QDB this ensured the BOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindLast;

    (FObjectFile.CurrentObj as TTestObject).Name := 'ObjecT0';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );

    (FObjectFile.CurrentObj as TTestObject).Name := 'ObjeCt1';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );


    { while testing with QDB this ensured the EOF property was not set before
      a FindNearest on the first object }
    FObjectFile.FindFirst;

    (FObjectFile.CurrentObj as TTestObject).Name := 'OBJect9';
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    { finding beyond the last object must position the last object }
    (FObjectFile.CurrentObj as TTestObject).Name := 'ObjEctZ';
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnAgeNameISAM;
begin
    if FObjectDatabase.Driver.SingleIndexOnly or
       not FObjectDatabase.Driver.IntegerIndexSupported then
        Exit;

    FObjectFile.IndexName := 'AgeName';

    (FObjectFile.CurrentObj as TTestObject).Age  := 106;
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
    Check( not FObjectFile.FindExact );

    (FObjectFile.CurrentObj as TTestObject).Age  := 5;
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object5';
    Check( FObjectFile.FindExact );
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Age  := 0;
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object5';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );


    (FObjectFile.CurrentObj as TTestObject).Age  := 5;
    (FObjectFile.CurrentObj as TTestObject).Name := '';
    FObjectFile.FindNearest;
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Age := 1;
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );


    (FObjectFile.CurrentObj as TTestObject).Age := 9;
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object9';
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    { finding beyond the last object must position the last object }
    (FObjectFile.CurrentObj as TTestObject).Age := 99;
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object9';
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    (FObjectFile.CurrentObj as TTestObject).Age      := 9;
    (FObjectFile.CurrentObj as TTestObject).Name     := 'Object99';
    (FObjectFile.CurrentObj as TTestObject).KeyValue := '99';
    FObjectFile.Insert;

    (FObjectFile.CurrentObj as TTestObject).Age := 8;
    (FObjectFile.CurrentObj as TTestObject).Name := 'ObjectZ';
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( FObjectFile.FindNext );
    Check((FObjectFile.CurrentObj as TTestObject).Age  = 9);
    Check((FObjectFile.CurrentObj as TTestObject).Name = 'Object99');
    FObjectFile.Delete;

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnAgeDescISAM;
begin
    if FObjectDatabase.Driver.SingleIndexOnly or
       not FObjectDatabase.Driver.IntegerIndexSupported or
       not FObjectDatabase.Driver.DescendingIndexSupported then
        Exit;

    FObjectFile.IndexName := 'AgeDesc';

    Check(FObjectFile.FindFirst);
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );

    Check(FObjectFile.FindLast);
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    (FObjectFile.CurrentObj as TTestObject).Age := 5;
    Check( FObjectFile.FindExact );

    FObjectFile.FindNext;
    CheckObjectValues( 4, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Age := 3;
    FObjectFile.FindNearest;
    CheckObjectValues( 3, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Age := 0;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Age := 10;
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnWeightNameISAM;
begin
    if FObjectDatabase.Driver.SingleIndexOnly or
       not FObjectDatabase.Driver.FloatIndexSupported then
        Exit;

    FObjectFile.IndexName := 'WName';

    (FObjectFile.CurrentObj as TTestObject).Weight  := 106.2;
    (FObjectFile.CurrentObj as TTestObject).Name    := 'Object1';
    Check( not FObjectFile.FindExact );

    { float values
    Object9 = -121.950
    Object8 = -108.400
    Object7 = -94.850
    Object6 = -81.300
    Object5 = -67.750
    Object4 = -54.200
    Object3 = -40.650
    Object2 = -27.100
    Object1 = -13.550
    }

    (FObjectFile.CurrentObj as TTestObject).Weight  := -123;
    (FObjectFile.CurrentObj as TTestObject).Name    := 'Object5';
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );

    (FObjectFile.CurrentObj as TTestObject).Weight := -68.5;
    (FObjectFile.CurrentObj as TTestObject).Name   := '';
    FObjectFile.FindNearest;
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Weight := -14;
    (FObjectFile.CurrentObj as TTestObject).Name   := 'Object9';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    { finding beyond the last object must position the last object }
    (FObjectFile.CurrentObj as TTestObject).Weight := 23.99;
    (FObjectFile.CurrentObj as TTestObject).Name   := '';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    (FObjectFile.CurrentObj as TTestObject).Weight   := -13.550;
    (FObjectFile.CurrentObj as TTestObject).Name     := 'Object11';
    (FObjectFile.CurrentObj as TTestObject).KeyValue := '11';
    FObjectFile.Insert;

    (FObjectFile.CurrentObj as TTestObject).Weight := -13.551;
    (FObjectFile.CurrentObj as TTestObject).Name := 'ObjectZ';
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( FObjectFile.FindNext );
    Check((FObjectFile.CurrentObj as TTestObject).Name = 'Object11');
    FObjectFile.Delete;

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnWeightAgeISAM;
begin
    if FObjectDatabase.Driver.SingleIndexOnly or
       not FObjectDatabase.Driver.IntegerIndexSupported or
       not FObjectDatabase.Driver.FloatIndexSupported then
        Exit;

    FObjectFile.IndexName := 'WAge';

    (FObjectFile.CurrentObj as TTestObject).Weight  := 106.2;
    (FObjectFile.CurrentObj as TTestObject).Name    := '5';
    Check( not FObjectFile.FindExact );

    { float values
    Object9 = -121.950
    Object8 = -108.400
    Object7 = -94.850
    Object6 = -81.300
    Object5 = -67.750
    Object4 = -54.200
    Object3 = -40.650
    Object2 = -27.100
    Object1 = -13.550
    }

    (FObjectFile.CurrentObj as TTestObject).Weight  := -123;
    (FObjectFile.CurrentObj as TTestObject).Age     := 5;
    FObjectFile.FindNearest;
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindPrior );

    (FObjectFile.CurrentObj as TTestObject).Weight := -68.5;
    (FObjectFile.CurrentObj as TTestObject).Age    := 0;
    FObjectFile.FindNearest;
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );

    (FObjectFile.CurrentObj as TTestObject).Weight := -14;
    (FObjectFile.CurrentObj as TTestObject).Age    := 9;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    { finding beyond the last object must position the last object }
    (FObjectFile.CurrentObj as TTestObject).Weight := 23.99;
    (FObjectFile.CurrentObj as TTestObject).Age    := 0;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( not FObjectFile.FindNext );

    (FObjectFile.CurrentObj as TTestObject).Name     := 'Object11';
    (FObjectFile.CurrentObj as TTestObject).Weight   := -13.550;
    (FObjectFile.CurrentObj as TTestObject).Age      := 11;
    (FObjectFile.CurrentObj as TTestObject).KeyValue := '11';
    FObjectFile.Insert;

    (FObjectFile.CurrentObj as TTestObject).Weight := -13.551;
    (FObjectFile.CurrentObj as TTestObject).Age    := 99;
    FObjectFile.FindNearest;
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    Check( FObjectFile.FindNext );
    Check((FObjectFile.CurrentObj as TTestObject).Age = 11);
    FObjectFile.Delete;

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestUpdateISAM;
begin
    Check( FObjectFile.FindFirst );
    (FObjectFile.CurrentObj as TTestObject).Name := 'ObjectA';
    (FObjectFile.CurrentObj as TTestObject).Data := 9213;
    FObjectFile.Update;

    { after the update, the first object in now Object2 }
    Check( FObjectFile.FindFirst );
    Check((FObjectFile.CurrentObj as TTestObject).Name = 'Object2');
    Check((FObjectFile.CurrentObj as TTestObject).Age  = 2);

    Check( FObjectFile.FindLast );
    Check((FObjectFile.CurrentObj as TTestObject).Name = 'ObjectA');
    Check((FObjectFile.CurrentObj as TTestObject).Data = 9213);

    { restore the first object }
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
    (FObjectFile.CurrentObj as TTestObject).Data := 0;
    FObjectFile.Update;

    { update without changing key }
    (FObjectFile.CurrentObj as TTestObject).Data := 9823;
    FObjectFile.Update;
    Check( FObjectFile.FindLast );
    Check( FObjectFile.FindFirst );
    Check((FObjectFile.CurrentObj as TTestObject).Data = 9823);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestDuplicateInsertISAM;
begin
    try
        (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
        FObjectFile.Insert;
    except
        Exit;
    end;
    Check( False, 'Duplicate insert should have raised exception');
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestDuplicateUpdateISAM;
begin
    FObjectFile.IndexName := 'Name';

    Check( FObjectFile.FindFirst );
    Check((FObjectFile.CurrentObj as TTestObject).Name = 'Object1');
    try
        (FObjectFile.CurrentObj as TTestObject).Name := 'Object2';
        FObjectFile.Update;
    except
        { Mar 11, 2001. MB Advantage local engine had problems here. An exception
          was raised, #7057, but the first object became 'Object2' even if an
          exception signaling the failure was raised. The workaround was to
          have the driver override CheckDuplicateUpdate.

          The same problem occured with FlashFiler.

          After checking in the Advantage table itself, the Name field is not
          update and does not violate the unique index, but the blob which contains
          the object is updated. This is probably the same problem with
          Flash Filer 2. This was confirmed as a problem on the Advantage.general
          newsgroup, but there was no response about a fix.
        }
        (*
        SendDebug( ' ' );
        if FObjectFile.FindFirst then
        repeat
            SendDebug( (FObjectFile.CurrentObj as TTestObject).Name );
        until not FObjectFile.FindNext;
        *)

        Exit;
    end;
    Check( False, 'Duplicate update should have raised exception');
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestDeleteObjectFile;
begin
    FObjectFile.IndexName := 'Name';
    Check(FObjectFile.FindFirst);
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    FObjectFile.Delete;
    CheckObjectValues( 2, FObjectFile.CurrentObj as TTestObject );
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
    Check(not FObjectFile.FindExact);

    (FObjectFile.CurrentObj as TTestObject).Name := 'Object3';
    Check(FObjectFile.FindExact);
    CheckObjectValues( 3, FObjectFile.CurrentObj as TTestObject );
    FObjectFile.Delete;
    CheckObjectValues( 4, FObjectFile.CurrentObj as TTestObject );
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object3';
    Check(not FObjectFile.FindExact);

    (FObjectFile.CurrentObj as TTestObject).Name := 'Object4';
    Check(FObjectFile.FindExact);
    CheckObjectValues( 4, FObjectFile.CurrentObj as TTestObject );
    FObjectFile.Delete;
    CheckObjectValues( 5, FObjectFile.CurrentObj as TTestObject );
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object4';
    Check(not FObjectFile.FindExact);

    Check(FObjectFile.FindLast);
    CheckObjectValues( 9, FObjectFile.CurrentObj as TTestObject );
    FObjectFile.Delete;
    CheckObjectValues( 8, FObjectFile.CurrentObj as TTestObject );
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object9';
    Check(not FObjectFile.FindExact);

    { Delete all items by repeatingly calling FindFirst. This will test
      FindFirst continues to loop until the file is empty }
    while FObjectFile.FindFirst do
        FObjectFile.Delete;
    Check( FObjectFile.IsEmpty );

    { add a few objects and testing deleting with FindLast }
    AddObjectISAM(1, FObjectFile);
    AddObjectISAM(2, FObjectFile);
    AddObjectISAM(3, FObjectFile);
    while FObjectFile.FindLast do
        FObjectFile.Delete;
    Check( FObjectFile.IsEmpty );
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestObjectFileHooks( const ObjectFile : TObjectFile );
begin
    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    ObjectFile.Open;
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeOpen') >= 0 );
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterOpen') >= 0 );

    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    AddObjectISAM(99, FObjectFile);
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeInsert') >= 0 );
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterInsert') >= 0 );
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeObjectPost') >= 0 );

    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    (ObjectFile.CurrentObj as TTestObject).Age := 32;
    ObjectFile.Update;
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeUpdate') >= 0 );
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterUpdate') >= 0 );
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeObjectPost') >= 0 );

    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    (ObjectFile.CurrentObj as TTestObject).Name := 'Object99';
    ObjectFile.Delete;
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeDelete') >= 0 );
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterDelete') >= 0 );

    AddObjectISAM(90, FObjectFile);
    AddObjectISAM(91, FObjectFile);

    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    Check(ObjectFile.FindFirst);
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterObjectRead') >= 0 );

    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    Check(ObjectFile.FindNext);
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterObjectRead') >= 0 );

    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    Check(ObjectFile.FindPrior);
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterObjectRead') >= 0 );

    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    Check(ObjectFile.FindLast);
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterObjectRead') >= 0 );

    while (ObjectFile.FindFirst) do
         ObjectFile.Delete;

    (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.Clear;
    ObjectFile.Close;
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('BeforeClose') >= 0 );
    Check( (ObjectFile.ORMapping as TTestObjectMapping).FiredMethods.IndexOf('AfterClose') >= 0 );
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestCreateDeleteFile;
begin
    { test CreateFile, DeleteFile backward compatibility methods }
    FObjectFile.CreateFile;
    FObjectFile.Open;
    AddObjectISAM(1, FObjectFile);
    Check(FObjectFile.FindFirst);
    CheckObjectValues( 1, FObjectFile.CurrentObj as TTestObject );
    FObjectFile.DeleteFile;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestObjectFileLockingAndPack;
var
    NewObjectFile : TObjectFile;
    Raised        : Boolean;
begin
    if not FObjectDatabase.Driver.FileLockingSupported then
        Exit;

    FObjectFile.Close;

    FObjectFile.OpenExclusive;

    NewObjectFile  := TObjectFile.Create( FObjectDatabase, 'Test2', TTestObjectMapping );
    try
        Raised := False;
        try
            NewObjectFile.Open;
        except
            Raised := True;
        end;
        if not Raised then
            raise Exception.Create('Exception should have been raised in TestObjectFileLockingAndPack');

    finally
        NewObjectFile.Free;
    end;

    { this is not an exhaustive test, but verifies that Pack does not fail }
    FObjectFile.Pack;

    FObjectFile.Close;
    FObjectFile.Open;

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTransactionsISAM;
var
    NewObjectFile : TObjectFile;
    TestObject    : TTestObject;

    procedure TestRollback;
    begin
        FObjectDatabase.StartTransaction;

        Check( FObjectFile.ObjectCount = 9 );

        FObjectFile.IndexName := 'Name';
        Check(FObjectFile.FindLast);
        (FObjectFile.CurrentObj as TTestObject).Age := 5021;
        FObjectFile.Update;
        Check(FObjectFile.FindPrior);
        Check(FObjectFile.FindNext);
        Check((FObjectFile.CurrentObj as TTestObject).Age = 5021);

        (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
        Check(FObjectFile.FindExact);
        FObjectFile.Delete;
        (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
        Check(not FObjectFile.FindExact);

        Check( FObjectFile.ObjectCount = 8 );


        TestObject := TTestObject.Create;
        NewObjectFile.CurrentObj := TestObject;
        TestObject.Free;
        (NewObjectFile.CurrentObj as TTestObject).Name := 'Object100';
        NewObjectFile.Insert;
        Check( NewObjectFile.ObjectCount = 1 );


        FObjectDatabase.RollBack;


        { bug in DbIsam which does not maintain ObjectCount after a roll back }
        if FDriverName <> 'TDBISAMOdbDriver' then
            Check( FObjectFile.ObjectCount = 9 );

        Check(FObjectFile.FindLast);
        Check((FObjectFile.CurrentObj as TTestObject).Age = 9);

        (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
        Check(FObjectFile.FindExact);

        { bug in DbIsam which does not maintain ObjectCount after a roll back }
        if FDriverName <> 'TDBISAMOdbDriver' then
            Check( NewObjectFile.ObjectCount = 0 );
        Check( not NewObjectFile.FindFirst );
    end;

    procedure TestCommit;
    begin
        FObjectDatabase.StartTransaction;

        Check( FObjectFile.ObjectCount = 9 );

        FObjectFile.IndexName := 'Name';
        Check(FObjectFile.FindLast);
        (FObjectFile.CurrentObj as TTestObject).Age := 5021;
        FObjectFile.Update;
        Check(FObjectFile.FindPrior);
        Check(FObjectFile.FindNext);
        Check((FObjectFile.CurrentObj as TTestObject).Age = 5021);

        (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
        Check(FObjectFile.FindExact);
        FObjectFile.Delete;
        (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
        Check(not FObjectFile.FindExact);

        Check( FObjectFile.ObjectCount = 8 );


        TestObject := TTestObject.Create;
        NewObjectFile.CurrentObj := TestObject;
        TestObject.Free;
        (NewObjectFile.CurrentObj as TTestObject).Name := 'Object100';
        NewObjectFile.Insert;
        Check( NewObjectFile.ObjectCount = 1 );

        FObjectDatabase.Commit;

        Check( FObjectFile.ObjectCount = 8 );

        Check(FObjectFile.FindLast);
        Check((FObjectFile.CurrentObj as TTestObject).Age  = 5021);
        Check((FObjectFile.CurrentObj as TTestObject).Name = 'Object9');
        { restore to previous value }
        (FObjectFile.CurrentObj as TTestObject).Age := 9;
        FObjectFile.Update;

        (FObjectFile.CurrentObj as TTestObject).Name := 'Object1';
        Check(not FObjectFile.FindExact);
        AddObjectISAM(1, FObjectFile);


        Check( NewObjectFile.FindFirst );
        Check((NewObjectFile.CurrentObj as TTestObject).Name = 'Object100');

    end;

begin
    if not FObjectDatabase.Driver.TransactionsSupported then
        Exit;

    { test transactions on two object files }
    NewObjectFile  := TObjectFile.Create( FObjectDatabase, 'Test2', TTestObjectMapping );
    try
        NewObjectFile.CreateDataset;
        NewObjectFile.Open;

        TestRollback;

        { MB Mar 11, 2001. Serious bug in Flash filer 2, on the second transaction
          the Insert in NewObjectFiles fails with a message indicating the "table
          is opened" ?!. The order of  TestCommit/TestRollback may be inverted,
          but the error  will occur on the second call. }
        TestCommit;

    finally
        NewObjectFile.DeleteDataset;
        NewObjectFile.Free;
    end;

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestObjectLockingObjectFile;
var
    NewObjectFile     : TObjectFile;
    NewObjectDatabase : TObjectDatabase;
    Raised            : Boolean;
begin
    if not FObjectDatabase.Driver.ObjectLockingSupported then
        Exit;

    { create the new file with a new database instance, to simulate as much
      as possible multiuser access }

    NewObjectDatabase := TObjectDatabase.Create( FDriverClass, FDatabaseName );
    NewObjectFile     := nil;
    try
        NewObjectDatabase.Open( FServerName, FUserName, FPassword );

        { MB Mar 26, 2001. Create the object file after opening the database,
          as some introspection methods return more accurate information
          once the database is opened. }
        NewObjectFile     := TObjectFile.Create( NewObjectDatabase, 'TestData', TTestObjectMapping );

        NewObjectFile.Open;
        NewObjectFile.IndexName := 'Name';

        FObjectFile.FindFirst;
        Check((FObjectFile.CurrentObj as TTestObject).Name = 'Object1');
        Check(FObjectFile.LockCurrent);

        NewObjectFile.FindFirst;
        Check((NewObjectFile.CurrentObj as TTestObject).Name = 'Object1');
        Check(not NewObjectFile.LockCurrent);

        Raised := False;
        try
            (NewObjectFile.CurrentObj as TTestObject).Age := 100;

            // Flash filer 2 allows deleting a locked record
            { don't test updating the locked record, as many drivers based
              on TDataset lock the record in the TDataset.Edit method, often
              causing the driver to first unlock the record in the same session.
              A good lock test is to attempt to delete the record }
            NewObjectFile.Delete;
        except
            on E:Exception do
            begin
                Raised := True;
            end;
        end;

        if not Raised then
            Check( False, 'Exception should have been raised in lock test');

        FObjectFile.UnLockCurrent;

        Check(NewObjectFile.LockCurrent);
        NewObjectFile.UnLockCurrent;

    finally
        NewObjectFile.Free;
        NewObjectDatabase.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestBookmarks;
var
    Bookmark : TOdbBookmark;
begin
    FObjectFile.IndexName := 'Name';
    (FObjectFile.CurrentObj as TTestObject).Name := 'Object4';
    Check(FObjectFile.FindExact);
    CheckObjectValues( 4, FObjectFile.CurrentObj as TTestObject );

    Bookmark := FObjectFile.GetBookmark;

    Check(FObjectFile.FindLast);
    Check((FObjectFile.CurrentObj as TTestObject).Name = 'Object9');

    if not FObjectDatabase.Driver.SingleIndexOnly and
       FObjectDatabase.Driver.IntegerIndexSupported then
        { test index restored on goto bookmark }
        FObjectFile.IndexName := 'Age';

    FObjectFile.GotoBookmark(Bookmark, True);
    CheckObjectValues( 4, FObjectFile.CurrentObj as TTestObject );
    Check(FObjectFile.IndexName = 'Name');
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestSQL;
var
    TestObject : TTestObject;
    Raised     : Boolean;
begin
    if not (damSQL in FObjectDatabase.Driver.GetDataAccessMethods) then
        Exit;

    Check( FObjectQuery.Database = FObjectDatabase );

    Check( AnsiUppercase(FObjectQuery.DatasetName) = AnsiUppercase('TestData'));
    Check( AnsiUppercase(FObjectQuery.DatasetName) = AnsiUppercase('TestData'));

    Raised := False;
    try
        Check( FObjectQuery.FieldByName('UnknownFld') = nil);
    except
        on E:EOdbFieldDoesNotExist do
            Raised := True
        else
            raise;
    end;
    if not Raised then
        Check(False, 'FieldByName failed');

    Check( FObjectQuery.Driver is FDriverClass);

    FObjectQuery.CreateDataset;
    TestObject := TTestObject.Create;
    try
        Check( not FObjectQuery.Opened );

        FObjectQuery.Select.Add('select Object from TestData');
        FObjectQuery.Open;
        try
            Check( FObjectQuery.IsEmpty );
            Check( FObjectQuery.Opened );
        finally
            FObjectQuery.Close;
            Check( not FObjectQuery.Opened );
        end;

        AddObjectSQL( 1, FObjectUpdate );
        AddObjectSQL( 2, FObjectUpdate );
        AddObjectSQL( 3, FObjectUpdate );
        AddObjectSQL( 4, FObjectUpdate );
        AddObjectSQL( 5, FObjectUpdate );
        AddObjectSQL( 6, FObjectUpdate );
        AddObjectSQL( 7, FObjectUpdate );
        AddObjectSQL( 8, FObjectUpdate );
        AddObjectSQL( 9, FObjectUpdate );


        TestTraverseByNameSQL;
        TestTraverseByAgeSQL;
        TestTraverseByWeightSQL;

        TestFindOnNameSQL;
        TestFindOnAgeSQL;
        TestFindOnWeightSQL;

        TestTransactionsSQL;

        TestUpdateSQL;
        TestDuplicateInsertSQL;
        TestDuplicateUpdateSQL;
        TestDeleteObjectQuery;



    finally
        TestObject.Free;

        { if the FObjectUpdate uses ISAM instead of SQL, it must be
          destroyed before deleting the dataset }
        FObjectUpdate.Free;
        FObjectUpdate := nil;

        FObjectQuery.DeleteDataset;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTraverseByNameSQL;
var
    Count : Integer;
begin
    FObjectQuery.Select.Clear;
    FObjectQuery.Select.Add( 'select Object,Name from TestData' );
    FObjectQuery.Select.Add( 'order by Name' );

    FObjectQuery.Open;
    try

        { check forward traversal }
        Check( FObjectQuery.FindFirst );
        CheckObjectValues( 1, FObjectQuery.CurrentObj as TTestObject );
        for Count := 2 to 9 do
        begin
            Check( FObjectQuery.FindNext );
            CheckObjectValues( Count, FObjectQuery.CurrentObj as TTestObject );
        end;
        Check( not FObjectQuery.FindNext );

        { check backward traversal }
        Check( FObjectQuery.FindLast );
        CheckObjectValues( 9, FObjectQuery.CurrentObj as TTestObject );
        for Count := 8 downto 1 do
        begin
            Check( FObjectQuery.FindPrior );
            CheckObjectValues( Count, FObjectQuery.CurrentObj as TTestObject );
        end;
        Check( not FObjectQuery.FindPrior );

    finally
        FObjectQuery.Close;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTraverseByAgeSQL;
var
    Count : Integer;
begin
    { Although SQL does not require the index, if integer indices are not
      supported, the Age field is not defined. }
    if not FObjectDatabase.Driver.IntegerIndexSupported then
        Exit;

    FObjectQuery.Select.Clear;
    FObjectQuery.Select.Add( 'select Object,Age from TestData' );
    FObjectQuery.Select.Add( 'order by Age' );

    FObjectQuery.Open;
    try

        { check forward traversal }
        Check( FObjectQuery.FindFirst );
        CheckObjectValues( 1, FObjectQuery.CurrentObj as TTestObject );
        for Count := 2 to 9 do
        begin
            Check( FObjectQuery.FindNext );
            CheckObjectValues( Count, FObjectQuery.CurrentObj as TTestObject );
        end;
        Check( not FObjectQuery.FindNext );

        { check backward traversal }
        Check( FObjectQuery.FindLast );
        CheckObjectValues( 9, FObjectQuery.CurrentObj as TTestObject );
        for Count := 8 downto 1 do
        begin
            Check( FObjectQuery.FindPrior );
            CheckObjectValues( Count, FObjectQuery.CurrentObj as TTestObject );
        end;
        Check( not FObjectQuery.FindPrior );

    finally
        FObjectQuery.Close;
    end;

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTraverseByWeightSQL;
var
    Count : Integer;
begin
    { Although SQL does not require the index, if float indices are not
      supported, the Weight field is not defined. }
    if not FObjectDatabase.Driver.FloatIndexSupported then
        Exit;


    FObjectQuery.Select.Clear;
    FObjectQuery.Select.Add( 'select Object,Weight from TestData' );
    FObjectQuery.Select.Add( 'order by Weight' );

    FObjectQuery.Open;
    try
        { check forward traversal. The weight value decreases as the other
          fields inrease. For example, the weight for 'Object1' is larger than
          the weight for 'Object2'. }
        Check( FObjectQuery.FindFirst );
        CheckObjectValues( 9, FObjectQuery.CurrentObj as TTestObject );
        for Count := 2 to 9 do
        begin
            Check( FObjectQuery.FindNext );
            CheckObjectValues( 10-Count, FObjectQuery.CurrentObj as TTestObject );
        end;
        Check( not FObjectQuery.FindNext );

        { check backward traversal }
        Check( FObjectQuery.FindLast );
        CheckObjectValues( 1, FObjectQuery.CurrentObj as TTestObject );
        for Count := 8 downto 1 do
        begin
            Check( FObjectQuery.FindPrior );
            CheckObjectValues( 10-Count, FObjectQuery.CurrentObj as TTestObject );
        end;
        Check( not FObjectQuery.FindPrior );

    finally
        FObjectQuery.Close;
    end;

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnNameSQL;
begin
    FObjectQuery.Select.Clear;
    FObjectQuery.Select.Add( 'select Object,Name from TestData' );
    FObjectQuery.Select.Add( 'where Name=''Object3''' );

    FObjectQuery.Open;
    try
        Check( FObjectQuery.FindFirst );
        CheckObjectValues( 3, FObjectQuery.CurrentObj as TTestObject );
        Check( not FObjectQuery.FindNext );

        Check( FObjectQuery.FindLast );
        CheckObjectValues( 3, FObjectQuery.CurrentObj as TTestObject );
        Check( not FObjectQuery.FindPrior );

    finally
        FObjectQuery.Close;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnAgeSQL;
begin
    FObjectQuery.Select.Clear;
    FObjectQuery.Select.Add( 'select Object,Age from TestData' );
    FObjectQuery.Select.Add( 'where Age=7' );

    FObjectQuery.Open;
    try
        Check( FObjectQuery.FindFirst );
        CheckObjectValues( 7, FObjectQuery.CurrentObj as TTestObject );
        Check( not FObjectQuery.FindNext );

        Check( FObjectQuery.FindLast );
        CheckObjectValues( 7, FObjectQuery.CurrentObj as TTestObject );
        Check( not FObjectQuery.FindPrior );

    finally
        FObjectQuery.Close;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestFindOnWeightSQL;
begin
    FObjectQuery.Select.Clear;
    FObjectQuery.Select.Add( 'select Object,Weight from TestData' );
    FObjectQuery.Select.Add( 'where (Weight>=-54.2001) and (Weight<=-54.1999)' );

    FObjectQuery.Open;
    try
        Check( FObjectQuery.FindFirst );
        CheckObjectValues( 4, FObjectQuery.CurrentObj as TTestObject );
        Check( not FObjectQuery.FindNext );

        Check( FObjectQuery.FindLast );
        CheckObjectValues( 4, FObjectQuery.CurrentObj as TTestObject );
        Check( not FObjectQuery.FindPrior );

    finally
        FObjectQuery.Close;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestTransactionsSQL;
var
    NewObjectUpdate : TObjectUpdate;
    TestObject      : TTestObject;
    FoundObj        : TTestObject;
    BdeAndLocal     : Boolean;

    procedure TestRollback;
    begin
        FObjectDatabase.StartTransaction;

        TestObject.Name     := 'Object9';
        TestObject.KeyValue := '9';
        TestObject.Age      := 5021;

        FObjectUpdate.Update(TestObject);

        { with IBX, the query is in another transaction, so it will not see
          the change. }
        if FDriverName <> 'TIbxOdbDriver' then
            Check((FObjectQuery.Search(TestObject) as TTestObject).Age = 5021);

        TestObject.Name := 'Object1';
        FObjectUpdate.Delete(TestObject);
        { with IBX, the query is in another transaction, so it will not see
          the change. }
        if FDriverName <> 'TIbxOdbDriver' then
            Check(FObjectQuery.Search(TestObject) = nil);
        TestObject.Name     := 'Object100';
        TestObject.KeyValue := '100';
        NewObjectUpdate.Insert(TestObject);

        { with IBX, the query is in another transaction, so it will not see
          the change. }
        if FDriverName <> 'TIbxOdbDriver' then
            Check(FObjectQuery.Search(TestObject) <> nil );
        FObjectDatabase.RollBack;

        TestObject.Name := 'Object100';
        Check(FObjectQuery.Search(TestObject) = nil );

        TestObject.Name := 'Object1';
        Check(FObjectQuery.Search(TestObject) <> nil);

        TestObject.Name := 'Object9';
        Check((FObjectQuery.Search(TestObject) as TTestObject).Age = 9);
    end;

    procedure TestCommit;
    begin
        FObjectDatabase.StartTransaction;

        TestObject.Name     := 'Object9';
        TestObject.Age      := 5021;
        TestObject.KeyValue := '9';
        FObjectUpdate.Update(TestObject);

        TestObject.Name := 'Object1';
        FObjectUpdate.Delete(TestObject);

        { With DbISAM, if Name and Key 100, a duplicate key exception is
          raised although, the previous insert has been rolled back }
        if FDriverName = 'TDbIsamOdbDriver' then
        begin
            TestObject.Name     := 'Object101';
            TestObject.KeyValue := '101';
        end
        else
        begin
            TestObject.Name     := 'Object100';
            TestObject.KeyValue := '100';
        end;
        NewObjectUpdate.Insert(TestObject);

        { with IBX, the query is in another transaction, so it will not see
          the change. }
        if FDriverName <> 'TIbxOdbDriver' then
            Check(FObjectQuery.Search(TestObject) <> nil );
        FObjectDatabase.Commit;

        TestObject.Name := 'Object9';
        Check((FObjectQuery.Search(TestObject) as TTestObject).Age = 5021);

        TestObject.Name := 'Object1';
        Check(FObjectQuery.Search(TestObject) = nil);

        if FDriverName = 'TDbIsamOdbDriver' then
            TestObject.Name := 'Object101'
        else
            TestObject.Name := 'Object100';
        Check(FObjectQuery.Search(TestObject) <> nil );

        { undo changes }
        if FDriverName = 'TDbIsamOdbDriver' then
            TestObject.Name := 'Object101'
        else
            TestObject.Name := 'Object100';
        FObjectUpdate.Delete(TestObject);

        AddObjectSQL(1, FObjectUpdate);

        TestObject.Name := 'Object9';
        FoundObj := FObjectQuery.Search(TestObject) as TTestObject;
        FoundObj.Age  := 9;
        FObjectUpdate.Update(FoundObj);
    end;

begin
    if not FObjectDatabase.Driver.TransactionsSupported then
        Exit;

    { test transactions with two object updates }
    NewObjectUpdate  := TObjectUpdate.Create( FObjectDatabase, 'TestData', TTestObjectMapping );
    TestObject       := TTestObject.Create;
    try
        { MB Mar 21, 2001. Rollback on paradox failed when using SQL query }
        BdeAndLocal := (FObjectDatabase.Driver is TBdeOdbDriver) and
                       not (FObjectDatabase.Driver as TBdeOdbDriver).SqlBased;
        if not BdeAndLocal then
            TestRollback;

        TestCommit;
    finally
        TestObject.Free;
        NewObjectUpdate.Free;
    end;

end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestUpdateSQL;
var
    TestObject : TTestObject;
begin
    TestObject := TTestObject.Create;
    try
        TestObject.Name      := 'Object3';
        TestObject.KeyValue  := '3';
        TestObject.Age       := 195;
        TestObject.Data      := 284;
        FObjectUpdate.Update(TestObject);

        with FObjectQuery.Search(TestObject) as TTestObject do
        begin
            Check( Age = 195 );
            Check( Data = 284 );
        end;
    finally
        TestObject.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestDuplicateInsertSQL;
var
    TestObject : TTestObject;
    Raised     : Boolean;
begin
    TestObject := TTestObject.Create;
    try
        TestObject.Name     := 'Object3';
        TestObject.KeyValue := '3';
        TestObject.Age      := 195;
        TestObject.Data     := 284;
        Raised := False;
        try
            FObjectUpdate.Insert(TestObject);
        except
            Raised := True;
        end;

        Check( Raised, 'Exception not raised in TestDuplicateInsertSQL');
    finally
        TestObject.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestDuplicateUpdateSQL;
var
    TestObject : TTestObject;
    Raised     : Boolean;
begin
    TestObject := TTestObject.Create;
    try
        TestObject.Name      := 'Object3';
        { this field was cause the duplicate key violation on update }
        TestObject.KeyValue  := '5';
        TestObject.Age       := 195;
        TestObject.Data      := 284;
        Raised := False;
        try
            FObjectUpdate.Update(TestObject);
        except
            Raised := True;
        end;

        Check( Raised, 'Exception not raised in TestDuplicateUpdateSQL');
    finally
        TestObject.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriverTest.TestDeleteObjectQuery;
var
    TestObject : TTestObject;
begin
    TestObject := TTestObject.Create;
    try
        TestObject.Name := 'Object3';
        FObjectUpdate.Delete(TestObject);

        Check( FObjectQuery.Search(TestObject) = nil );

        AddObjectSQL(3, FObjectUpdate);

    finally
        TestObject.Free;
    end;
end;

{--------------------------------------------------------------------------}


end.
